from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.scene.sequence import ChunkSequence


def setup_tiles(ctx: AlabamaContext, sequence: ChunkSequence):
    """
    Sets up the tiles based on the resolution
    stolen from the one and only autocompressor.com's source code.
    🤑
    """
    if ctx.prototype_encoder.get_codec() == Codec.av1 and not ctx.multi_res_pipeline:
        if (
            ctx.prototype_encoder.tile_cols == -1
            and ctx.prototype_encoder.tile_rows == -1
        ):
            width, height = ctx.get_output_res()

            def calculate_tiles(video_width, video_height, target_pixels_per_tile):
                # when target_pixels_per_tile is 2,000,000 -> 1 tile at 1080p
                # when target_pixels_per_tile is 1,000,000 -> 2 tiles at 1080p
                rowsl = 0
                colsl = 0
                ctpx = video_width * video_height
                # current tile pixels, starts at the full size of the video
                # we subdivide until <4/3 of targetPixelsPerTile
                ctar = video_width / video_height
                # current tile aspect ratio, we subdivide into cols if >1 and rows if ≤1
                while ctpx >= target_pixels_per_tile * 4 / 3:
                    if ctar > 1:
                        # Subdivide into columns, add 1 to colsl and halve ctar, then halve ctpx
                        colsl += 1
                        ctar /= 2
                        ctpx /= 2
                    else:
                        # Subdivide into rows, add 1 to rowsl and double ctar, then halve ctpx
                        rowsl += 1
                        ctar *= 2
                        ctpx /= 2
                return {
                    "tileRowsLog2": rowsl,
                    "tileColsLog2": colsl,
                    "tileRows": 2**rowsl,
                    "tileCols": 2**colsl,
                }

            tile_info = calculate_tiles(width, height, 1_666_666)
            ctx.log(
                f"Calculated literal tile cols: {tile_info['tileCols']}; literal tile rows: {tile_info['tileRows']}",
                category="analyzing_content_logs"
            )
            ctx.prototype_encoder.tile_rows = tile_info["tileRowsLog2"]
            ctx.prototype_encoder.tile_cols = tile_info["tileColsLog2"]

    return ctx



import json
import os
from datetime import datetime
from typing import List

import pandas as pd
from pandas import DataFrame
from tqdm import tqdm

from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.impl.vmaf import VmafOptions
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.scene.chunk import ChunkObject

only_one = False


def get_test_files(test_files_count: int = -1):
    # get test_files env
    test_files = os.environ.get("TEST_FILES")
    if test_files is None:
        test_files = "/mnt/data/objective-1-fast"

    # get all files in dir
    files = []
    for root, dirs, filenames in os.walk(test_files):
        for f in filenames:
            files.append(os.path.join(root, f))

    # filter out non y4m
    files = [f for f in files if f.endswith(".y4m")]

    if test_files_count != -1:
        # limit so no less than zero and no more than the number of files is picked
        test_files_count = max(min(test_files_count, len(files)), 0)
        files = files[:test_files_count]

    return files


def get_test_env() -> str:
    pwd = os.getcwd()
    date = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    test_env = pwd + f"/run_{date}/"
    # if not os.path.exists(test_env):
    #     os.mkdir(test_env)

    return test_env


def run_test(
    enc: Encoder, input_file: str, version: str, test_env: str = get_test_env()
) -> EncodeStats:
    # get the basename of the file
    basename = os.path.basename(input_file).split(".")[0]

    # create and set an output input_path for the video file
    # eg ./encode_test_2021-03-30_16-00-00/STARCRAFT_60f_420control.ivf
    output_path = f"{test_env}{basename}_{version}{enc.get_chunk_file_extension()}"

    threads = os.cpu_count()

    # set the input and output input_path
    enc.chunk = ChunkObject(path=input_file)
    enc.output_path = output_path
    enc.threads = threads
    stats = enc.run(
        override_if_exists=False,
        metric_to_calculate=Metric.VMAF,
        calcualte_ssim=True,
        metric_params=VmafOptions(neg=True, threads=threads),
    )
    stats.version = version
    stats.basename = basename
    return stats


def run_tests(
    enc: Encoder,
    version: str = "control",
    test_env: str = get_test_env(),
    test_files_count: int = -1,
) -> List[EncodeStats]:
    """
    Run a test on all the files in the TEST_FILES env
    :param test_env: where to save the encoded files
    :param enc: Configured encoder object to run the test with
    :param version: version name to attach to the stats, e.g., "control"
    :return: a list of EncodeStats objects
    """
    test_files = get_test_files(test_files_count)

    stat_reports = []

    for input_path in tqdm(test_files, desc=f"Running tests for {version}", leave=None):
        stat_reports.append(
            run_test(enc, input_file=input_path, version=version, test_env=test_env)
        )

    return stat_reports


def save_stats(
    stat_reports: List[EncodeStats],
    experiment_name: str = "Experiment",
    test_env: str = get_test_env(),
):
    """
    Save the stats to a json file
    :param test_env: where to save the encoded files
    :param stat_reports: a list of EncodeStats objects
    :param experiment_name: name of the experiment
    :return: None
    """
    # save the stats to a json file
    output_path = f"{test_env}{experiment_name}.json"
    if os.path.exists(output_path):
        print(f"WARNING: {output_path} already exists, overwriting")
    with open(output_path, "w") as f:
        # save as a list of dicts
        json.dump([s.get_dict() for s in stat_reports], f)


def report(stat_reports: List[EncodeStats], experiment_name: str = "Experiment"):
    print(f"## {experiment_name}")

    class StatHolder:
        def __init__(
            self, basename, sts: List[EncodeStats], control: EncodeStats = None
        ):
            self.basename = basename
            self.sts = sts
            self.control = control

    holders = []

    # put stats with the same basename together in StatHolder, also put control in the holder
    for stat in stat_reports:
        if stat is None:
            raise ValueError("stat is None")
        found = False
        for s in holders:
            if s.basename == stat.basename:
                if stat.version == "control":
                    s.control = stat
                else:
                    s.sts.append(stat)
                found = True
                break
        if not found:
            if stat.version == "control":
                holders.append(StatHolder(stat.basename, [], stat))
            else:
                holders.append(StatHolder(stat.basename, [stat]))

    # print the stats
    for holder in holders:
        print(
            f"| _{holder.basename}_ |  time taken  | kpbs | vmaf  | BD Change % | time Change % | performance |\n"
            "|----------|:------:|:----:|:-----:|:-----------:|:---:|:---:|"
        )

        control = EncodeStats() if holder.control is None else holder.control

        # since we will sometimes be having multiple controls and dont set the control object, we have a fallback
        # to just print them
        stat_to_process = [*holder.sts, holder.control]

        if holder.control is None:
            stat_to_process = holder.sts

        for stat in stat_to_process:
            stat.time_encoding = round(stat.time_encoding, 2)
            curr_db_rate = stat.bitrate / stat.vmaf

            if control.bitrate == -1 or control.vmaf == -1:
                db_change_from_control = 0
            else:
                db_change_from_control = (
                    (curr_db_rate - control.bitrate / control.vmaf)
                    / (control.bitrate / control.vmaf)
                    * 100
                )
                db_change_from_control = round(db_change_from_control, 2)

            if control.time_encoding == -1:
                time_change_from_control = 0
            else:
                time_change_from_control = (
                    0
                    if control.time_encoding == -1
                    else (
                        (stat.time_encoding - control.time_encoding)
                        / control.time_encoding
                        * 100
                    )
                )
                time_change_from_control = round(time_change_from_control, 2)

            preformance = (stat.vmaf / stat.time_encoding / stat.bitrate) * 1000

            print(
                f"| {stat.version} |"
                f" {stat.time_encoding}s |"
                f" {stat.bitrate} "
                f"| {round(stat.vmaf, 2)} | "
                f" {db_change_from_control}% |"
                f" {time_change_from_control}% |"
                f" {round(preformance, 2)} |"
            )


def run_tests_across_range(
    encs: List[Encoder],
    title: str,
    test_env: str = get_test_env(),
    crfs: List[int] = [
        16,
        20,
        24,
        28,
        33,
        39,
        48,
        51,
        57,
        60,
    ],
    bitrates: List[int] = [250, 500, 1000, 2500, 4000, 8000],
    test_files_count: int = -1,
    skip_vbr: bool = False,
) -> None:
    """
    Run tests across a range of crfs and bitrates using different encoders
    :param encs: 0 is control, rest are tests
    :param title: title of the experiment
    :param test_env: where to save the encoded files
    :param crfs: crfs to test across
    :param bitrates: bitrates to test across
    :return: None
    """
    stats_crf = []
    stats_bitrates = []

    if len(encs) == 0:
        raise ValueError("encs must have at least 1 encoder")

    for enc in tqdm(encs, desc=f"Running tests for {title}", leave=None):
        enc_index = encs.index(enc)

        enc_env = test_env + f"{enc_index}/"

        if not os.path.exists(enc_env):
            os.mkdir(enc_env)

        if not skip_vbr:
            bitrate_env = enc_env + "bitrate/"
            if not os.path.exists(bitrate_env):
                os.mkdir(bitrate_env)

            for bitrate in tqdm(bitrates, desc="Bitrates", leave=None):
                enc.bitrate = bitrate
                enc.passes = 3
                enc.rate_distribution = EncoderRateDistribution.VBR
                version = f"enc{enc_index}_{bitrate}kbs"
                if enc_index == 0:
                    version = f"control_{bitrate}kbs"
                stats_bitrates += run_tests(
                    enc,
                    version,
                    test_env=bitrate_env,
                    test_files_count=test_files_count,
                )

        crf_env = enc_env + "crf/"
        if not os.path.exists(crf_env):
            os.mkdir(crf_env)

        for crf in tqdm(crfs, desc="CRF", leave=None):
            enc.crf = crf
            enc.passes = 1
            enc.rate_distribution = EncoderRateDistribution.CQ

            version = f"enc{enc_index}_{crf}crf"
            if enc_index == 0:
                version = f"control_{crf}crf"
            stats_crf += run_tests(
                enc, version, test_env=crf_env, test_files_count=test_files_count
            )

    report(stats_crf, experiment_name=f"{title} CRF")
    save_stats(stats_crf, experiment_name=f"{title} CRF", test_env=test_env)

    if not skip_vbr:
        report(stats_bitrates, experiment_name=f"{title} Bitrates")
        save_stats(
            stats_bitrates, experiment_name=f"{title} Bitrates", test_env=test_env
        )


def read_report(report_path: str) -> DataFrame:
    with open(report_path) as f:
        data = json.load(f)

    # only get  size bitrate vmaf ssim vmaf_percentile_1 vmaf_percentile_5 vmaf_percentile_10 vmaf_percentile_25 vmaf_percentile_50 vmaf_avg basename version
    for i in range(len(data)):
        data[i] = {
            k: data[i][k]
            for k in (
                "size",
                "bitrate",
                "vmaf",
                "ssim",
                "vmaf_percentile_1",
                "vmaf_percentile_5",
                "vmaf_percentile_10",
                "vmaf_percentile_25",
                "vmaf_percentile_50",
                "vmaf_avg",
                "basename",
                "version",
                "time_encoding",
            )
        }

    df = pd.DataFrame(data)

    df["rate"] = (
        df["version"]
        .str.split("_")
        .str[-1]
        .str.replace("crf", "")
        .str.replace("kbs", "")
        .astype(int)
    )
    df["test_group"] = df["version"].str.split("_").str[0]

    return df


if __name__ == "__main__":
    # test_env = get_test_env()
    # experiment_name = "Testing speed presets, 12 is control"
    #
    # enc = AbstractEncoderSvtenc()
    # enc.update(speed=12, crf=30, passes=1, rate_distribution=RateDistribution.CQ)
    # enc.threads = 12
    #
    # stats = run_tests(enc, version="control", test_env=test_env)
    #
    # enc.speed = 4
    #
    # stats += run_tests(enc, "speed4", test_env=test_env)
    #
    # enc.speed = 8
    #
    # stats += run_tests(enc, "speed8", test_env=test_env)
    #
    # enc.speed = 2
    #
    # stats += run_tests(enc, "speed2", test_env=test_env)
    #
    # save_stats(stats, experiment_name=experiment_name, test_env=test_env)
    #
    # report(stat_reports=stats, experiment_name=experiment_name)
    pass



import os
import re

from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.util.bin_utils import get_binary


class ImageMetrics:
    @staticmethod
    def butteraugli_score(reference_img_path: str, distorted_img_path: str):
        if not os.path.exists(reference_img_path) or not os.path.exists(
            distorted_img_path
        ):
            raise FileNotFoundError("images not found, please check the paths")

        cli = f'{get_binary("butteraugli")} "{reference_img_path}"  "{distorted_img_path}"'
        try:
            result_string = run_cli(cli).get_output()
            # if the result does not contain a single number, then it failed
            if not re.search(r"[\d.]+", result_string):
                raise AttributeError

            return float(result_string)
        except AttributeError or ValueError:
            raise RuntimeError(f"Failed getting butteraugli, CLI: {cli}")

    @staticmethod
    def psnr_score(reference_img_path, distorted_img_path):
        cli = (
            f"{get_binary('ffmpeg')} -hide_banner -i {reference_img_path} -i {distorted_img_path} "
            f"-filter_complex psnr -f null -"
        )

        result_string = run_cli(cli).get_output()
        try:
            # Regex to get average score out of:
            # [Parsed_psnr_0 @ 0x55f6705fa200] PSNR y:49.460782 u:52.207017 v:51.497660 average:50.118351
            # min:48.908778 max:51.443411

            match = re.search(r"average:([\d.]+)", result_string)
            psnr_score = float(match.group(1))
            return psnr_score
        except AttributeError:
            print(
                f"Failed getting psnr comparing {reference_img_path} against {distorted_img_path}"
            )
            print(cli)
            return 0

    @staticmethod
    def ssim_score(reference_img_path, distorted_img_path):
        cli = f" ffmpeg -hide_banner -i {reference_img_path} -i {distorted_img_path} -filter_complex ssim -f null -"

        result_string = run_cli(cli).get_output()
        try:
            match = re.search(r"All:\s*([\d.]+)", result_string)
            ssim_score = float(match.group(1))
            return ssim_score
        except AttributeError:
            print(
                f"Failed getting ssim comparing {reference_img_path} against {distorted_img_path}"
            )
            print(cli)
            return 0

    @staticmethod
    def vmaf_score(reference_img_path, distorted_img_path):
        cli = f" ffmpeg -hide_banner -i {reference_img_path} -i {distorted_img_path} -lavfi libvmaf -f null -"

        result_string = run_cli(cli).get_output()
        try:
            match = re.compile(r"VMAF score: ([0-9]+\.[0-9]+)").search(result_string)
            vmaf_score = float(match.group(1))
            return vmaf_score
        except AttributeError:
            print(
                f"Failed getting vmath comparing {reference_img_path} against {distorted_img_path} command:"
            )
            print(cli)
            return 0



from alabamaEncode.conent_analysis.chunk.final_encode_step import (
    FinalEncodeStep,
)
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.scene.chunk import ChunkObject


class PlainFinalEncode(FinalEncodeStep):
    def run(
        self, enc: Encoder, chunk: ChunkObject, ctx: AlabamaContext, encoded_a_frame
    ) -> EncodeStats:
        metric, _ = ctx.get_metric_target()
        stats = enc.run(
            metric_to_calculate=metric if ctx.calc_final_vmaf else None,
            metric_params=ctx.get_vmaf_options(),
            on_frame_encoded=encoded_a_frame,
        )
        if metric == Metric.VMAF and ctx.calc_final_vmaf:
            if stats.metric_results.frames is not None:
                vmaf_frame_scores = {}
                for frame in stats.metric_results.frames:
                    vmaf_frame_scores[frame["frameNum"]] = frame["metrics"]["vmaf"]
                ctx.get_kv().set(
                    "vmaf_frame_scores",
                    chunk.chunk_index,
                    vmaf_frame_scores,
                    individual_mode=True,
                )

        return stats

    def dry_run(self, enc: Encoder, chunk: ChunkObject) -> str:
        joined = " && ".join(enc.get_encode_commands())
        return joined



import json
import os

from tqdm import tqdm

from alabamaEncode.conent_analysis.chunk.final_encode_steps.dynamic_target_vmaf import (
    DynamicTargetVmaf,
)
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.core.util.timer import Timer
from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.parallel_execution.command import BaseCommandObject
from alabamaEncode.scene.chunk import ChunkObject


class ChunkEncoder(BaseCommandObject):
    """
    Take a chunk, run it through the pipeline and encode it
    """

    def __init__(self, ctx: AlabamaContext, chunk: ChunkObject):
        super().__init__()
        self.ctx = ctx
        self.chunk = chunk
        self.encoded_a_frame_callback: callable = None
        self.pin_to_core = -1
        # how long (seconds) before we time out the final encoding
        # currently set to 30 minutes
        self.final_encode_timeout = 1800
        self.run_on_celery = False

    def supports_encoded_a_frame_callback(self):
        return isinstance(self.ctx.prototype_encoder, EncoderSvt) and not isinstance(
            self.ctx.chunk_encode_class, DynamicTargetVmaf
        )

    def run(self) -> [int, EncodeStats]:
        timeing = Timer()

        timeing.start("chunk")
        try:
            timeing.start("analyze_step")

            enc = self.ctx.get_encoder()
            enc.pin_to_core = self.pin_to_core
            enc.chunk = self.chunk
            for step in self.ctx.chunk_analyze_chain:
                timeing.start(f"analyze_step_{step.__class__.__name__}")
                enc = step.run(self.ctx, self.chunk, enc)
                timeing.stop(f"analyze_step_{step.__class__.__name__}")

            rate_search_time = timeing.stop("analyze_step")

            enc.running_on_celery = self.run_on_celery

            if self.ctx.dry_run:
                print(f"dry run chunk: {self.chunk.chunk_index}")
                print(self.ctx.chunk_encode_class.dry_run(enc, self.chunk))
                return

            timeing.start("final_step")
            chunk_stats = self.ctx.chunk_encode_class.run(
                enc,
                chunk=self.chunk,
                ctx=self.ctx,
                encoded_a_frame=self.encoded_a_frame_callback,
            )
            timeing.stop("final_step")
        except Exception as e:
            tqdm.write(f"{self.chunk.log_prefix()}encoding failed: {e}")
            if os.path.exists(self.chunk.chunk_path):
                os.remove(self.chunk.chunk_path)
            return

        valid = self.chunk.verify_integrity(
            length_of_sequence=self.ctx.total_chunks, quiet=True
        )
        self.ctx.get_kv().set("chunk_integrity", self.chunk.chunk_index, not valid)

        timeing.stop("chunk")
        timing_stats = timeing.finish()

        if chunk_stats is None:
            return self.pin_to_core, None

        chunk_stats.total_fps = round(
            self.chunk.get_frame_count() / (timing_stats["chunk"]), 2
        )
        chunk_stats.chunk_index = self.chunk.chunk_index
        chunk_stats.rate_search_time = rate_search_time

        with open(f"{self.ctx.temp_folder}/chunks.log", "a") as f:
            f.write(json.dumps(chunk_stats.__dict__()) + "\n")

        self.ctx.get_kv().set(
            "chunk_timing", self.chunk.chunk_index, timing_stats, individual_mode=True
        )

        return self.pin_to_core, chunk_stats.__dict__()



from typing import List

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.path import PathAlabama
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution


class EncoderX264(Encoder):
    def get_pretty_name(self) -> str:
        return "X264"

    def get_codec(self) -> Codec:
        return Codec.h264

    def get_version(self) -> str:
        # x264 core:164 r3095 baee400
        # Syntax: x264 [options] -o outfile infile
        return (
            run_cli(f"{get_binary('x264')} --help")
            .get_output()
            .split("\n")[0]
            .replace("x264", "")
            .strip()
        )

    def get_encode_commands(self) -> List[str]:
        self.speed = max(min(self.speed, 9), 0)

        if not self.hdr:
            self.bit_override = 8

        taskset = ""

        if self.pin_to_core != -1:
            taskset += f" taskset -a -c {self.pin_to_core} "

        kommand = f" {self.get_ffmpeg_pipe_command()} |{taskset}{get_binary('x264')} - --stdin y4m "

        kommand += f" --threads {self.threads} "

        kommand += f" --output-depth {self.bit_override} "
        if self.hdr:
            kommand += f" --profile high10 --hdr "
            colormatrix = self.matrix_coefficients
            if colormatrix == "bt2020-ncl":
                colormatrix = "bt2020nc"
            chromeloc = self.chroma_sample_position
            if chromeloc == "topleft":
                chromeloc = "0"
            kommand += (
                f" --colorprim {self.color_primaries} --transfer {self.transfer_characteristics} "
                f"--colormatrix {colormatrix} --chromaloc {chromeloc} "
                f"--cll {self.maximum_content_light_level},{self.maximum_frame_average_light_level}"
            )
            # if self.svt_master_display:
            #     kommand += f' --mastering-display "{self.svt_master_display}" '
        else:
            kommand += " --colorprim bt709 --transfer bt709 --colormatrix bt709 "
            kommand += " --profile high --force-cfr "

        match self.rate_distribution:
            case EncoderRateDistribution.CQ:
                self.passes = 1
                kommand += f" --crf {self.crf}"
            case EncoderRateDistribution.VBR:
                kommand += f" --bitrate {self.bitrate} "
            case EncoderRateDistribution.CQ_VBV:
                kommand += f" --crf {self.crf} --bitrate {self.bitrate} "
            case EncoderRateDistribution.VBR_VBV:
                raise Exception("FATAL: rate distribution VBR_VBV not supported")

        if self.x264_ipratio != -1:
            kommand += f" --ipratio {self.x264_ipratio} "

        if self.x264_pbratio != -1:
            kommand += f" --pbratio {self.x264_pbratio} "

        if self.x264_aq_strength != -1:
            kommand += f" --aq-strength {self.x264_aq_strength} "

        if self.x264_merange != -1:
            kommand += f" --merange {self.x264_merange} "

        if self.x264_bframes != -1:
            kommand += f" --bframes {self.x264_bframes} "

        if self.x264_rc_lookahead != -1:
            kommand += f" --rc-lookahead {self.x264_rc_lookahead} "

        if self.x264_ref != -1:
            kommand += f" --ref {self.x264_ref} "

        if self.x264_me != "":
            kommand += f" --me {self.x264_me} "

        if self.x264_subme != -1:
            kommand += f" --subme {self.x264_subme} "

        if self.x264_non_deterministic:
            kommand += " --non-deterministic "

        if self.x264_vbv_maxrate != -1:
            kommand += f" --vbv-maxrate {self.x264_vbv_maxrate} "

        if self.x264_vbv_bufsize != -1:
            kommand += f" --vbv-bufsize {self.x264_vbv_bufsize} "

        if self.x264_slow_firstpass:
            kommand += " --slow-firstpass "

        if not self.x264_mbtree:
            kommand += " --no-mbtree "

        match self.speed:
            case 9:
                kommand += " --preset ultrafast"
            case 8:
                kommand += " --preset superfast"
            case 7:
                kommand += " --preset veryfast"
            case 6:
                kommand += " --preset faster"
            case 5:
                kommand += " --preset fast"
            case 4:
                kommand += " --preset medium"
            case 3:
                kommand += " --preset slow"
            case 2:
                kommand += " --preset slower"
            case 1:
                kommand += " --preset veryslow"
            case 0:
                kommand += " --preset placebo"

        if self.x264_tune != "":
            kommand += f" --tune {self.x264_tune} "

        kommand += f" --keyint 240 "
        kommand += f" --min-keyint 240 "

        kommand += " --stitchable "

        kommand += f" --muxer mkv "

        kommand += f" --fps {Ffmpeg.get_fps_fraction(PathAlabama(self.chunk.path))} "

        if self.passes > 1 or self.x264_collect_pass or self.passes <= -2:
            kommand += f' --stats "{self.output_path}.stats" '

        if self.passes == 1 and self.x264_collect_pass:
            kommand += f" --pass 1 "

        if self.passes == 1:
            return [f'{kommand} -o "{self.output_path}"']
        elif self.passes > 1:
            com = []
            for i in range(self.passes):
                com += [f'{kommand} --pass {i + 1} -o "{self.output_path}"']

            com += [
                f"rm -f {self.output_path}.stats",
                f"rm -f {self.output_path}.stats.mbtree",
            ]

            return com
        elif self.passes == -2:
            # if passes == -2, do second pass only
            com = [f'{kommand} --pass 2 -o "{self.output_path}"']
            com += [
                f"rm -f {self.output_path}.stats",
                f"rm -f {self.output_path}.stats.mbtree",
            ]
            return com
        elif self.passes == -3:
            # if passes == -3, do second and third pass
            com = [
                f'{kommand} --pass 2 -o "{self.output_path}"',
                f'{kommand} --pass 3 -o "{self.output_path}"',
            ]
            com += [
                f"rm -f {self.output_path}.stats",
                f"rm -f {self.output_path}.stats.mbtree",
            ]
            return com

    def get_chunk_file_extension(self) -> str:
        return ".mkv"

    def supports_float_crfs(self) -> bool:
        return True

    # def parse_output_for_output(self, buffer) -> List[str]:
    #     # 42 frames: 3.60 fps, 2481.60 kb/s
    #     if buffer is None:
    #         return []
    #     match = re.search(r"\d+ frames: .+ kb/s", buffer)
    #     if match:  # check if we are past the header, also extract the string
    #         _match = re.search(
    #             r"(\d+) frames: ([0-9.]+) fps, ([0-9.]+) kb/s",
    #             match.group(0),
    #         )  # parse out the frame number, time, and bitrate
    #         return [_match.group(1), _match.group(2), _match.group(3)]
    #     else:
    #         return []



import os

import numpy as np
from matplotlib import pyplot as plt

from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.scene.sequence import ChunkSequence


def plot_vmaf(ctx: AlabamaContext, sequence: ChunkSequence, show_bpp=False):
    print("Plotting vmaf...")
    vmaf_scores = ctx.get_kv().get_all("vmaf_frame_scores")
    crfs = ctx.get_kv().get_all("target_vmaf")

    bpp = {}
    pixels_in_frame = sequence.chunks[0].width * sequence.chunks[0].height
    for chunk in sequence.chunks:
        chunk_size_bits = os.path.getsize(chunk.chunk_path) * 8
        bpp[chunk.chunk_index] = (
            chunk_size_bits / pixels_in_frame / chunk.get_frame_count()
        )

    all_scores = []
    for chunk_scores in vmaf_scores.values():
        all_scores.extend(chunk_scores.values())

    total_frames = sum([len(chunk_scores) for chunk_scores in vmaf_scores.values()])
    # Define a space per frame (e.g., 0.01 inch per frame)
    space_per_frame = 0.01

    # Calculate figure width based on total frames and space per frame
    fig_width = total_frames * space_per_frame

    # cap at 1000
    fig_width = min(fig_width, 100)

    # scale dpi based on width
    dpi = 100
    if fig_width > 10:
        dpi = 200

    fig_height = 8

    # make height proportional to width
    fig_height = fig_height * (fig_width / 10)

    plt.figure(figsize=(fig_width, fig_height), dpi=dpi)

    plt.plot(all_scores, label="VMAF", color="steelblue")
    plt.ylabel("VMAF Score")
    plt.ylim(0, 100)

    plt.axhline(ctx.vmaf, color="g", linestyle="--", label="Target VMAF")

    crf_values = []
    for chunk_scores in vmaf_scores.values():
        chunk_num = list(vmaf_scores.keys())[
            list(vmaf_scores.values()).index(chunk_scores)
        ]
        crf_values.extend([crfs[chunk_num]] * len(chunk_scores))

    ax1 = plt.gca()
    ax2 = plt.twinx()
    ax2.plot(crf_values, color="darkorange", linestyle="--", label="CRF")
    ax2.set_ylabel("CRF")
    ax2.set_ylim(0, 63)

    if show_bpp:
        # Create a third axis
        ax3 = ax1.twinx()

        # Calculate bpp values for each chunk
        bpp_values = [bpp[int(chunk_num)] for chunk_num in vmaf_scores.keys()]

        # Extend bpp values for each chunk by the number of frames in that chunk
        extended_bpp_values = []
        for chunk_num, bpp_value in zip(vmaf_scores.keys(), bpp_values):
            chunk_frame_count = sequence.chunks[int(chunk_num)].get_frame_count()
            extended_bpp_values.extend([bpp_value] * chunk_frame_count)

        # Plot extended bpp values on the third axis
        ax3.plot(extended_bpp_values, color="purple", linestyle="--", label="BPP")
        ax3.set_ylabel("BPP (rescaled)")

    mean_vmaf = np.mean(all_scores)
    median_vmaf = np.median(all_scores)
    harmonic_mean = 1 / np.mean([1 / x for x in all_scores if x != 0])

    plt.text(
        0.02,
        0.10,
        f"Mean: {mean_vmaf:.2f}\nMedian: {median_vmaf:.2f}\nHarmonic mean: {harmonic_mean}\nTarget: {ctx.vmaf}",
        transform=ax1.transAxes,
        fontsize=9,
        verticalalignment="top",
        bbox=dict(boxstyle="round", facecolor="wheat", alpha=0.5),
    )

    plt.xlabel("Frame")

    if show_bpp:
        plt.title("VMAF Scores, CRFs and BPP")
        lines1, labels1 = ax1.get_legend_handles_labels()
        lines2, labels2 = ax2.get_legend_handles_labels()
        lines3, labels3 = ax3.get_legend_handles_labels()
        ax2.legend(
            lines1 + lines2 + lines3, labels1 + labels2 + labels3, loc="lower right"
        )
    else:
        plt.title("VMAF against CRF")
        lines1, labels1 = ax1.get_legend_handles_labels()
        lines2, labels2 = ax2.get_legend_handles_labels()
        ax2.legend(lines1 + lines2, labels1 + labels2, loc="lower right")

    plt.savefig(f"{ctx.output_folder}/vmaf_plot.svg")
    plt.close()



import os

from alabamaEncode.core.util.bin_utils import register_bin
from alabamaEncode.encoder.impl.X264 import EncoderX264
from alabamaEncode.experiments.util.ExperimentUtil import get_test_files
from alabamaEncode.scene.chunk import ChunkObject

if __name__ == "__main__":
    test_env = "./tst_streaming_output/"
    test_env = os.path.abspath(test_env)
    if not os.path.exists(test_env):
        os.mkdir(test_env)
    register_bin("SvtAv1EncApp", "/home/kokoniara/.local/opt/SvtAv1EncApp")
    enc = EncoderX264()
    enc.chunk = ChunkObject(path=get_test_files()[0])
    enc.output_path = os.path.join(test_env, "output.mkv")
    enc.speed = 12
    enc.crf = 30
    enc.passes = 1

    enc.run(
        on_frame_encoded=lambda frame, bitrate, fps: print(f"{frame} {bitrate} {fps}")
    )



import os

from alabamaEncode.conent_analysis.chunk.final_encode_step import FinalEncodeStep
from alabamaEncode.conent_analysis.opinionated_vmaf import get_vmaf_list
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.scene.chunk import ChunkObject


class EncodeMultiResFinals(FinalEncodeStep):
    def run(
        self, enc: Encoder, chunk: ChunkObject, ctx: AlabamaContext, encoded_a_frame
    ) -> EncodeStats:
        if ctx.get_kv().get("multires_final_paths", "final_paths") is None:
            return None
        # a = {
        #     "res": "720",
        #     "chunks": [
        #         {"chunk_index": "0", "crf": "11"},
        #         {"chunk_index": "1", "crf": "12"},
        #     ],
        # }
        #

        for vmaf in get_vmaf_list(enc.get_codec()):
            obj = ctx.get_kv().get("multires_trellis", str(vmaf))
            res = obj["res"] + ":-2"
            for c in obj["chunks"]:
                if c["chunk_index"] == chunk.chunk_index:
                    enc.crf = c["crf"]
                    break

            # set res
            vf = ctx.prototype_encoder.video_filters.split(",")
            scale_str = f"scale={res}:flags=lanczos"

            # replace the first scale filter if it exists
            for i in range(len(vf)):
                if ":flags=lanczos" in vf[i]:
                    vf[i] = scale_str
                    break

            # if there is no scale filter, add it to the end
            if scale_str not in ",".join(vf):
                vf.append(scale_str)

            vf = [f for f in vf if f != ""]

            enc.video_filters = ",".join(vf)

            enc.output_path = f"{ctx.temp_folder}{chunk.chunk_index}_{vmaf}{enc.get_chunk_file_extension()}"

            if not os.path.exists(enc.output_path):
                ctx.get_kv().set(
                    "multires_final_paths",
                    f"{chunk.chunk_index}_{vmaf}",
                    enc.output_path,
                )
                enc.run()

        return None

    def dry_run(self, enc: Encoder, chunk: ChunkObject) -> str:
        raise Exception(f"dry_run not implemented for {self.__class__.__name__}")



import multiprocessing
import os
import re

from celery.worker.autoscale import Autoscaler as CeleryAutoscaler

from alabamaEncode.core.util.cli_executor import run_cli


# https://gist.github.com/hussainfolio3/c5246f59f9e5c31fa720524fb45b2077
# never go above one worker = one core
# if we have 12 cores then max is 12 workers


def parse_swap_usage() -> float:
    """
    Parse the output of 'swapon -s' and return the swap usage in percent.
    :return:
    """
    # Execute the 'swapon -s' and get the output
    output = run_cli("swapon -s").get_output()

    # Split the output into lines
    lines = output.split("\n")[1:]  # We skip the header

    for line in lines:
        if line:  # Ignore empty lines
            parts = line.split()

            # We assume that the line structure is correct here,
            # i.e., there are at least 5 parts.
            filename, type_, size, used, priority = (
                parts[0],
                parts[1],
                parts[2],
                parts[3],
                parts[4],
            )

            # Calculate and print usage Proportion
            usage = int(used) / int(size)
            # print(f"Swap usage for {filename}: {usage * 100:.2f}%")
            return usage * 100


class Load:
    def __init__(self):
        self.num_cpus = multiprocessing.cpu_count()

    def get_load(self):
        """Get cpu load in %"""
        load1min, load5min, load15min = os.getloadavg()
        # Prevent divide by zero
        if load1min < 0.001:
            load1min = 0.001
        return 1.0 * load1min / self.num_cpus

    re_total = re.compile(r"MemTotal:\s+(?P<total>\d+)\s+kB")
    re_free = re.compile(r"MemFree:\s+(?P<free>\d+)\s+kB")

    def get_free_mem(self):
        """Return percentage of free memory 0.0 to 1.0."""
        # If not, make it work for most linux distros.
        with open("/proc/meminfo", "rb") as f:
            mem = f.read().decode("utf-8")
            return (
                1.0
                * int(self.re_free.search(mem).group("free"))
                / int(self.re_total.search(mem).group("total"))
            )


# Load tests
if __name__ == "__main__":
    load = Load()
    print(f"Load: {load.get_load()}")
    print(f"Free mem: {load.get_free_mem()}")
    print(f"Swap usage: {parse_swap_usage()}%")


class DAAutoscaler(CeleryAutoscaler):
    # Try to keep the load above this point.
    LOAD_MIN = 0.8
    # Try to keep the load below this.
    LOAD_MAX = 1.1
    # We need this percentage of free memory to scale up.
    MEM_FREE_SCALE_UP = 0.3
    # Any less than this memory and we scale down.
    MEM_FREE_SCALE_DOWN = 0.1

    MAX_SCALE_UP = 5
    MIN_SCALE_DOWN = 1

    def __init__(self, *args, **kwargs):
        self.load = Load()
        print(f"DAAutoscaler: Num CPUs {self.load.num_cpus}")
        super(DAAutoscaler, self).__init__(*args, **kwargs)

    def _maybe_scale(self, req=None):
        """Scale up or down if we too much/little load or memory."""
        cur_load = self.load.get_load()
        mem_free = self.load.get_free_mem()

        if cur_load < self.LOAD_MIN and mem_free > self.MEM_FREE_SCALE_UP:
            mul = int(self.LOAD_MAX / cur_load)
            mul = max(min(mul, self.MAX_SCALE_UP), self.MIN_SCALE_DOWN)
            print(
                f"DAAutoscaler: Scale Up {mul}X {round(cur_load, 2)} free={round(100 * mem_free, 2)}%"
            )
            self.scale_up(mul)
            return True
        if cur_load > self.LOAD_MAX or mem_free < self.MEM_FREE_SCALE_DOWN:
            mul = int(cur_load / self.LOAD_MAX)
            print(
                f"DAAutoscaler: Scale Down {mul}X {round(cur_load, 2)} free={round(100 * mem_free, 2)}%"
            )
            self.scale_down(mul)
            return True

        print(f"DAAutoscaler: Ok {cur_load} {100 * mem_free}%")



import copy
import os
import shutil

from alabamaEncode.conent_analysis.chunk.chunk_analyse_step import (
    ChunkAnalyzePipelineItem,
)
from alabamaEncode.conent_analysis.opinionated_vmaf import (
    get_crf_limits,
    get_vmaf_probe_speed,
    get_vmaf_probe_offset,
)
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.calculate import get_metric_from_stats
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.scene.chunk import ChunkObject


class TargetVmaf(ChunkAnalyzePipelineItem):
    def run(self, ctx: AlabamaContext, chunk: ChunkObject, enc: Encoder) -> Encoder:
        enc_copy = copy.deepcopy(enc)

        kv = ctx.get_kv()

        crf_from_kv = kv.get(bucket="target_vmaf", key=str(chunk.chunk_index))
        if crf_from_kv is not None:
            if enc.supports_float_crfs():
                enc.crf = float(crf_from_kv)
            else:
                enc.crf = int(crf_from_kv)
            return enc

        metric, target_metric = ctx.get_metric_target()

        probe_file_base = ctx.get_probe_file_base(chunk.chunk_path)

        def get_score(_crf):
            kv_key = f"{chunk.chunk_index}_{_crf}"
            result_from_kv = kv.get(bucket="target_vmaf_probes", key=kv_key)
            if result_from_kv is not None:
                return float(result_from_kv)

            enc_copy.crf = _crf
            enc_copy.output_path = os.path.join(
                probe_file_base,
                f"probe.{_crf}{enc_copy.get_chunk_file_extension()}",
            )
            enc_copy.speed = max(get_vmaf_probe_speed(enc_copy, ctx), enc.speed)
            enc_copy.override_flags = ""
            # TODO: calculate metrics outside enc.run to add the flexibility to calc other ones
            stats: EncodeStats = enc_copy.run(
                metric_to_calculate=metric,
                metric_params=ctx.get_vmaf_options(),
                override_if_exists=False,
            )

            # TODO: offset the faster preset by metric amount
            result = get_metric_from_stats(
                stats=stats,
                statistical_representation=ctx.vmaf_target_representation,
            )
            if metric == Metric.VMAF:
                result += get_vmaf_probe_offset(enc_copy)

            kv.set(bucket="target_vmaf_probes", key=kv_key, value=result)
            return result

        probes = ctx.probe_count
        if probes > 3:
            ctx.log(
                f"probe count in TargetVmaf is set to {probes}, this is too high, <=3",
                level=1,
            )

        trys = []
        low_crf, high_crf = get_crf_limits(enc_copy, ctx)
        depth = 0
        mid_crf = 0
        while low_crf <= high_crf and depth < probes:
            mid_crf = (low_crf + high_crf) // 2

            if (depth == 2 and ctx.probe_count == 3) or (
                depth == 1 and ctx.probe_count == 2
            ):
                ll, lh = get_crf_limits(enc_copy, ctx)
                m = (ll + lh) // 2
                # if closer to edge then the middle, use that edge
                if abs(mid_crf - ll) < abs(mid_crf - m):
                    mid_crf = ll
                else:
                    mid_crf = lh

                ctx.log(
                    f"{chunk.log_prefix()} skipping to crf edge: {mid_crf}",
                    category="probe",
                )

            # don't try the same crf twice
            if mid_crf in [t[0] for t in trys]:
                break

            statistical_representation = get_score(mid_crf)

            ctx.log(
                f"{chunk.log_prefix()} crf: {mid_crf} {metric.name}: {statistical_representation} "
                f"attempt {depth + 1}/{probes}",
                category="probe",
            )

            if statistical_representation > target_metric:
                low_crf = mid_crf + 1
            else:
                high_crf = mid_crf - 1

            trys.append((mid_crf, statistical_representation))
            depth += 1

        # To limit the overhead, we do only 2-3 binary search probes.
        # It's too shallow to get a good result on its own,
        # that's why we can pick two close points and interpolate between them.
        # This improves prediction by a lot while only needing at ~3 probes
        if len(trys) > 1:
            # sort by metric difference from target
            points = sorted(trys, key=lambda _x: abs(_x[1] - target_metric))

            # get the two closest points
            crf_low, metric_low = points[0]
            crf_high, metric_high = points[1]

            # means multiple probes got the same score, aka all back screens etc
            if metric_high - metric_low == 0:
                crf = mid_crf
            else:
                crf = crf_low + (crf_high - crf_low) * (
                    (target_metric - metric_low) / (metric_high - metric_low)
                )

                if not enc_copy.supports_float_crfs():
                    crf = int(crf)

                crf_min, crf_max = get_crf_limits(enc_copy, ctx)

                crf = max(min(crf, crf_max), crf_min)
        else:
            crf = mid_crf

        ctx.log(f"{chunk.log_prefix()}Decided on crf: {crf}", category="probe")

        kv.set(bucket="target_vmaf", key=str(chunk.chunk_index), value=crf)

        # clean up probe folder
        if os.path.exists(probe_file_base):
            shutil.rmtree(probe_file_base)

        enc.crf = crf

        return enc



import json
import os
from statistics import mean

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.bin_utils import register_bin
from alabamaEncode.core.util.cli_executor import run_cli_parallel
from alabamaEncode.core.util.path import PathAlabama
from alabamaEncode.metrics.exception import VmafException
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.metrics.options import MetricOptions
from alabamaEncode.metrics.result import MetricResult
from alabamaEncode.scene.chunk import ChunkObject


class VmafOptions(MetricOptions):
    def __init__(
        self,
        phone=False,
        uhd=False,
        neg=False,
        no_motion=False,
        subsample=-1,
        **kwargs,
    ):
        self.phone = phone
        self.uhd = uhd
        self.neg = neg
        self.no_motion = no_motion
        self.subsample = subsample
        super().__init__(**kwargs)

    def get_model(self) -> str or None:
        models = get_models()
        if self.no_motion:
            if not os.path.exists(models["normal_neg_nomotion"]):
                return None
            return f'path={models["normal_neg_nomotion"]}'
        if self.neg:
            if self.uhd:
                if not os.path.exists(models["uhd_neg"]):
                    return None
                return f'path={models["uhd_neg"]}'
            if self.phone:
                if not os.path.exists(models["normal_neg"]):
                    return None
                return f'path={models["normal_neg"]}:name=phonevmaf:enable_transform'
            else:
                if not os.path.exists(models["normal_neg"]):
                    return None
                return f'path={models["normal_neg"]}'

        else:
            if self.uhd:
                if not os.path.exists(models["uhd"]):
                    return None
                return f'path={models["uhd"]}'
            if self.phone:
                if not os.path.exists(models["normal"]):
                    return None
                return f'path={models["normal"]}:name=phonevmaf:enable_transform'
            else:
                if not os.path.exists(models["normal"]):
                    return None
                return f'path={models["normal"]}'

    @staticmethod
    def test():
        print(VmafOptions().get_model())
        assert VmafOptions(uhd=True).get_model() != VmafOptions().get_model()
        assert VmafOptions(phone=True).get_model() != VmafOptions().get_model()
        assert VmafOptions(neg=True).get_model() != VmafOptions().get_model()


def calc_vmaf(
    chunk: ChunkObject,
    vmaf_options: VmafOptions,
    log_path="",
):
    assert vmaf_options is not None

    from alabamaEncode.metrics.calculate import get_input_pipes

    owo = get_input_pipes(chunk=chunk, options=vmaf_options)

    ref_pipe = owo["ref_pipe"]
    dist_pipe = owo["dist_pipe"]
    ref_command = owo["ref_command"]
    dist_command = owo["dist_command"]

    if log_path == "":
        log_path = f"/tmp/{os.path.basename(chunk.chunk_path)}.vmaflog"

    vmaf_model = vmaf_options.get_model()

    vmaf_command = (
        f'{get_binary("vmaf")} -q --json --output "{log_path}" '
        f"--reference {ref_pipe} "
        f"--distorted {dist_pipe}"
        f" --threads {vmaf_options.threads}"
    )

    if vmaf_model is not None:
        vmaf_command += f" --model {vmaf_model}"

    if vmaf_options.subsample != -1 and vmaf_options.subsample is not None:
        vmaf_command += f" --subsample {vmaf_options.subsample}"

    cli_results = run_cli_parallel(
        [
            ref_command,
            dist_command,
            vmaf_command,
        ]
    )

    from alabamaEncode.metrics.calculate import cleanup_input_pipes

    cleanup_input_pipes(owo)

    for c in cli_results:
        try:
            c.verify()
        except RuntimeError as e:
            raise VmafException(
                f"Could not run vmaf command: {e}, {[c.output for c in cli_results]}"
            )

    try:
        log_decoded = json.load(open(log_path))
    except (json.decoder.JSONDecodeError, FileNotFoundError):
        raise VmafException(
            f"Could not decode vmaf log: {log_path}, {[c.output for c in cli_results]}"
        )

    os.remove(log_path)

    result = VmafResult(
        pooled_metrics=log_decoded["pooled_metrics"],
        _frames=log_decoded["frames"],
        fps=log_decoded["fps"],
    )
    return result


class VmafResult(MetricResult):
    def __init__(self, _frames=None, pooled_metrics=None, fps=None):
        if pooled_metrics is None:
            pooled_metrics = {}

        self.fps = fps
        self.frames = _frames

        if _frames is not None:
            frames = []
            for frame in _frames:
                if "vmaf" in frame["metrics"]:
                    frames.append([frame["frameNum"], frame["metrics"]["vmaf"]])
                else:
                    frames.append([frame["frameNum"], frame["metrics"]["phonevmaf"]])
            frames.sort(key=lambda x: x[0])
            # calc 1 5 10 25 50 percentiles
            vmaf_scores = [x[1] for x in frames]
            vmaf_scores.sort()
            self.percentile_1 = vmaf_scores[int(len(vmaf_scores) * 0.01)]
            self.percentile_5 = vmaf_scores[int(len(vmaf_scores) * 0.05)]
            self.percentile_10 = vmaf_scores[int(len(vmaf_scores) * 0.1)]
            self.percentile_25 = vmaf_scores[int(len(vmaf_scores) * 0.25)]
            self.percentile_50 = vmaf_scores[int(len(vmaf_scores) * 0.5)]

            if "vmaf" in pooled_metrics:
                self.mean = pooled_metrics["vmaf"]["mean"]
                self.harmonic_mean = pooled_metrics["vmaf"]["harmonic_mean"]
            else:
                self.mean = mean([x[1] for x in frames])
                self.harmonic_mean = 1 / mean([1 / x[1] for x in frames if x[1] != 0])

            self.max = max([x[1] for x in frames])
            self.min = min([x[1] for x in frames])

            self.std_dev = 0
            for frame in frames:
                self.std_dev += (frame[1] - self.mean) ** 2
            self.std_dev /= len(frames)
            self.std_dev **= 0.5

    def __str__(self):
        return f"{self.mean}"

    def __repr__(self):
        return (
            f"VmafResult(mean={self.mean},"
            f" fps={self.fps},"
            f" harmonic_mean={self.harmonic_mean},"
            f" prct_1={self.percentile_1},"
            f" prct_5={self.percentile_5},"
            f" std_dev={self.std_dev})"
        )


if __name__ == "__main__":
    from alabamaEncode.metrics.calculate import calculate_metric

    register_bin("vmaf", "/home/kokoniara/.local/opt/vmaf")

    reference_path = PathAlabama("/mnt/data/objective-1-fast/MINECRAFT_60f_420.y4m")
    distorted_path = PathAlabama(
        "/mnt/data/objective-1-fast/minecruft_test_encode_av1.mkv"
    )

    print("calcing phone vmaf with 1080p comparison display")
    result: VmafResult = calculate_metric(
        reference_path=reference_path,
        distorted_path=distorted_path,
        options=VmafOptions(phone=True, threads=10),
        metric=Metric.VMAF,
    )
    print(result.mean)
    print(result.harmonic_mean)
    print(result.percentile_1)
    # etc with different statistical representations
    print(result.__repr__())

    print("calcing normal vmaf with 1080p comparison display")
    a = calculate_metric(
        reference_path=reference_path,
        distorted_path=distorted_path,
        options=VmafOptions(threads=10),
        metric=Metric.VMAF,
    )
    print(a.__repr__())

    print("calcing uhd vmaf with 4k comparison display")
    a = calculate_metric(
        reference_path=reference_path,
        distorted_path=distorted_path,
        options=VmafOptions(uhd=True, threads=10),
        metric=Metric.VMAF,
    )
    print(a.__repr__())

    print("Using chunk object, 720p reference display and mobile model")
    chunk = ChunkObject(path=reference_path.path)
    chunk.chunk_path = distorted_path.path
    a = calculate_metric(
        chunk=chunk,
        options=VmafOptions(phone=True, threads=10),
        metric=Metric.VMAF,
    )
    print(a.__repr__())


model_links = [
    [
        "https://github.com/Netflix/vmaf/raw/master/model/vmaf_4k_v0.6.1.json",
        "vmaf_4k_v0.6.1.json",
    ],
    [
        "https://github.com/Netflix/vmaf/raw/master/model/vmaf_v0.6.1.json",
        "vmaf_v0.6.1.json",
    ],
    [
        "https://github.com/Netflix/vmaf/raw/master/model/vmaf_4k_v0.6.1neg.json",
        "vmaf_4k_v0.6.1neg.json",
    ],
    [
        "https://github.com/Netflix/vmaf/raw/master/model/vmaf_v0.6.1neg.json",
        "vmaf_v0.6.1neg.json",
    ],
    [
        "https://raw.githubusercontent.com/kingstefan26/alabamaEncoder/master/vmaf_v0.6.1neg-nomotion.json",
        "vmaf_v0.6.1neg-nomotion.json",
    ],
]


def download_vmaf_models():

    models_dir = os.path.expanduser("~/.alabamaEncoder/vmaf_models")
    if not os.path.exists(models_dir):
        os.makedirs(models_dir)

    try:
        import requests

        for link in model_links:
            if not os.path.exists(os.path.join(models_dir, link[1])):
                print("Downloading VMAF model")
                r = requests.get(link[0], allow_redirects=True)
                open(os.path.join(models_dir, link[1]), "wb").write(r.content)

        for link in model_links:
            if not os.path.exists(os.path.join(models_dir, link[1])):
                print(f"Something went wrong accessing {link[1]}")
    except Exception as e:
        print(f"Failed downloading VMAF models, using built-in ones as fallback, {e}")


def get_models() -> dict[str, str]:
    models_dir = os.path.expanduser("~/.alabamaEncoder/vmaf_models")

    # turn the model paths into absolute paths
    for link in model_links:
        link[1] = os.path.join(models_dir, link[1])

    model_dict = {
        "uhd": model_links[0][1],
        "normal": model_links[1][1],
        "uhd_neg": model_links[2][1],
        "normal_neg": model_links[3][1],
        "normal_neg_nomotion": model_links[4][1],
    }

    return model_dict



import os
import re
import time
from typing import List

from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.util.path import PathAlabama
from alabamaEncode.scene.chunk import ChunkObject
from alabamaEncode.scene.sequence import ChunkSequence


def do_autocrop(ctx: AlabamaContext, sequence: ChunkSequence):
    """
    Does autocrop if needed
    """
    if ctx.auto_crop and ctx.crop_string == "":
        output = ctx.get_kv().get_global("autocrop")
        if output is None:
            start = time.time()
            print("Running cropdetect...")
            output = do_cropdetect(ctx.input_file)
            print(f"Computed crop: {output} in {int(time.time() - start)}s")
            path = PathAlabama(ctx.input_file)
            out_path = PathAlabama(ctx.output_file)

            def gen_prew(ss, i) -> List[str]:
                _p = f'"{out_path.get()}.{i}.cropped.jpg"'
                run_cli(
                    f"{get_binary('ffmpeg')} -v error -y -ss {ss} -i {path.get_safe()} "
                    f"-vf crop={output} -vframes 1 {_p}"
                ).verify()
                p2 = f'"{out_path.get()}.{i}.uncropped.jpg"'
                run_cli(
                    f"{get_binary('ffmpeg')} -v error -y -ss {ss} -i {path.get_safe()} -vframes 1 {p2}"
                )
                return [_p.replace('"', ""), p2.replace('"', "")]

            if not ctx.auto_accept_autocrop:
                print("Creating previews")
                generated_paths = gen_prew(60, 0)
                generated_paths += gen_prew(120, 1)
                print(
                    "Created crop previews in output folder, if you want to use this crop,"
                    " click enter, type anything to abort"
                )

                if input() != "":
                    print("Aborting")
                    for p in generated_paths:
                        if os.path.exists(p):
                            os.remove(p)
                    quit()

                for p in generated_paths:
                    if os.path.exists(p):
                        os.remove(p)

            ctx.get_kv().set_global("autocrop", output)

        vf = ctx.prototype_encoder.video_filters.split(",")
        for i in range(len(vf)):
            if "crop=" in vf[i]:
                vf[i] = f"crop={output}"
        if "crop=" not in ",".join(vf):
            # add to the start
            vf.insert(0, f"crop={output}")

        ctx.prototype_encoder.video_filters = ",".join(vf)

    return ctx


if __name__ == "__main__":
    print("AUTOCROP IMPLEMENTATION TEST")
    ctx = AlabamaContext()
    ctx.auto_crop = True
    ctx.input_file = "/mnt/data/downloads/Foundation.S02E08.2160p.WEB.h265-ETHEL[TGx]/Foundation.S02E08.2160p.WEB.h265-ETHEL.mkv"
    # ctx.input_file = "/mnt/data/downloads/Halo.S01.2160p.UHD.BluRay.Remux.HDR.DV.HEVC.Atmos-PmP/Halo.S01E04.Homecoming.2160p.UHD.BluRay.Remux.HDR.DV.HEVC.Atmos-PmP.mkv"
    ctx.output_file = "/home/kokoniara/showsEncode/HALO (2022)/s1/e4/out.webm"
    ctx = do_autocrop(ctx, None)
    print("Crop string: ", ctx.crop_string)
    assert (
        ctx.crop_string == "3840:1920:0:120"
    ), f"Expected 3840:1920:0:120 got {ctx.crop_string}"


def do_cropdetect(path: str = None):
    alabama = PathAlabama(path)
    alabama.check_video()
    fps = Ffmpeg.get_video_frame_rate(alabama)
    length = Ffmpeg.get_video_length(alabama) * fps

    #  create a 10-frame long chunk at 20% 40% 60% 80% of the length
    probe_chunks = []
    for i in range(0, 100, 20):
        first_frame_index = int(length * (i / 100))
        probe_chunks.append(
            ChunkObject(
                path=path,
                last_frame_index=first_frame_index + 10,
                first_frame_index=first_frame_index,
                framerate=fps,
            )
        )

    # function that takes a chunk and output its first cropdetect
    def get_crop(chunk) -> str:
        try:
            return re.search(
                r"-?\d+:-?\d+:-?\d+:-?\d+",
                run_cli(
                    f"{get_binary('ffmpeg')} -hwaccel auto {chunk.get_ss_ffmpeg_command_pair()} -vframes 10 -vf "
                    f"cropdetect -f null -"
                ).get_output(),
            ).group(0)
        except AttributeError:
            return ""

    # get the crops
    crops = [get_crop(chunk) for chunk in probe_chunks]

    # out of the 5 crops, get the most common crop
    most_common_crop = max(set(crops), key=crops.count)

    return most_common_crop



from abc import ABC, abstractmethod

from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.scene.chunk import ChunkObject


class ChunkAnalyzePipelineItem(ABC):
    """
    Sets up an Encoder for the final encoding, sometimes does analysis to find the parameters but doesn't have to
    """

    @abstractmethod
    def run(
        self,
        ctx,
        chunk: ChunkObject,
        enc: Encoder,
    ) -> Encoder:
        pass



from enum import Enum


class Codec(Enum):
    av1: int = 0
    h264: int = 1
    h265: int = 2
    vp8: int = 3
    vp9: int = 4

    def __str__(self):
        return self.name

    @staticmethod
    def from_str(codec_name: str):
        if codec_name == "av1":
            return Codec.av1
        elif codec_name == "h264":
            return Codec.h264
        elif codec_name == "h265":
            return Codec.h265
        elif codec_name == "vp8":
            return Codec.vp8
        elif codec_name == "vp9":
            return Codec.vp9
        else:
            raise ValueError(f"FATAL: Unknown codec name: {codec_name}")

    @staticmethod
    def get_all():
        return [Codec.av1, Codec.h264, Codec.h265, Codec.vp8, Codec.vp9]



from enum import Enum


class EncoderRateDistribution(Enum):
    VBR: int = 0  # variable bitrate
    CQ: int = 1  # constant quality/crf
    CQ_VBV: int = 2  # constant quality with limited bitrate
    VBR_VBV: int = 3  # variable bitrate with buffer to limit bitrate



from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.scene.sequence import ChunkSequence


def tune_args_for_fdlty_or_apl(ctx: AlabamaContext, sequence: ChunkSequence):
    match ctx.args_tune:
        case "fidelity":
            ctx.log("Tuning for fidelity", category="analyzing_content_logs")
            if ctx.simple_denoise:
                for _ in range(3):
                    ctx.log(
                        "YOU ARE USING SIMPLE DENOISE, THIS IS NOT RECOMMENDED FOR FIDELITY TUNING",
                        category="analyzing_content_logs"
                    )
            ctx.prototype_encoder.svt_tune = 1
            ctx.prototype_encoder.svt_tf = 0
            ctx.prototype_encoder.svt_enable_variance_boost = 1

        case "appeal":
            ctx.log("Tuning for appeal", category="analyzing_content_logs")
            ctx.prototype_encoder.qm_max = 8
            ctx.prototype_encoder.qm_min = 0
            ctx.prototype_encoder.svt_tune = 0
            ctx.prototype_encoder.svt_enable_variance_boost = 0
        case "balanced":
            ctx.log("Tuning for balanced appeal and fidelity")
            ctx.prototype_encoder.svt_tune = 0
        case _:
            raise RuntimeError(f"Invalid args_tune: {ctx.args_tune}")

    if isinstance(ctx.prototype_encoder, EncoderSvt) and ctx.prototype_encoder.is_psy():
        ctx.log("Found svt psy, using tune 3", category="analyzing_content_logs")
        ctx.prototype_encoder.svt_tune = 3

    return ctx



import json

from alabamaEncode.metrics.result import MetricResult


class EncodeStats:
    """
    Stats class for encoders
    """

    def __init__(
        self,
        time_encoding: int = -1,
        bitrate: int = -1,
        vmaf: float = -1,
        ssim: float = -1,
        ssim_db: float = -1,
        size: int = -1,
        total_fps: int = -1,
        target_miss_proc: int = -1,
        rate_search_time: int = -1,
        chunk_index: int = -1,
        basename: str = "",
        version: str = "",
        metric_result: MetricResult = None,
        length_frames: int = -1,
    ):
        self.time_encoding = time_encoding
        self.bitrate = bitrate
        self.vmaf = vmaf
        self.ssim = ssim
        self.ssim_db = ssim_db
        self.size = size  # size in kilo Bytes
        self.total_fps = total_fps
        self.target_miss_proc = target_miss_proc
        self.rate_search_time = rate_search_time
        self.chunk_index = chunk_index
        self.basename = basename
        self.version = version
        self.metric_results = metric_result or MetricResult()
        self.length_frames = length_frames

    def __dict__(self):
        return {
            "time_encoding": self.time_encoding,
            "bitrate": self.bitrate,
            "chunk_index": self.chunk_index,
            "length_frames": self.length_frames,
            "metric": self.metric_results.mean,
            "metric_class": self.metric_results.__class__.__name__,
            "ssim": self.ssim,
            "ssim_db": self.ssim_db,
            "size": self.size,
            "total_fps": self.total_fps,
            "target_miss_proc": self.target_miss_proc,
            "rate_search_time": self.rate_search_time,
            "metric_percentile_1": self.metric_results.percentile_1,
            "metric_percentile_5": self.metric_results.percentile_5,
            "metric_percentile_10": self.metric_results.percentile_10,
            "metric_percentile_25": self.metric_results.percentile_25,
            "metric_percentile_50": self.metric_results.percentile_50,
            "metric_avg": self.metric_results.mean,
            "basename": self.basename,
            "version": self.version,
        }

    def save(self, path):
        """
        Save stats to a file
        :param path: json file path
        :return: nun
        """
        with open(path, "w") as f:
            json.dump(self.__dict__(), f)

    def __str__(self):
        return str(self.__dict__)



def parse_rd(ctx):
    """
    Sets up the rate distribution mode
    """
    if ctx.bitrate_string is not None and ctx.bitrate_string != "":
        if "auto" in ctx.bitrate_string or "-1" in ctx.bitrate_string:
            ctx.find_best_bitrate = True
        else:
            if "M" in ctx.bitrate_string or "m" in ctx.bitrate_string:
                ctx.bitrate_string = ctx.bitrate_string.replace("M", "")
                ctx.prototype_encoder.bitrate = (
                    int(ctx.prototype_encoder.bitrate) * 1000
                )
            else:
                ctx.bitrate_string = ctx.bitrate_string.replace("k", "")
                ctx.bitrate_string = ctx.bitrate_string.replace("K", "")

                ctx.prototype_encoder.bitrate = int(ctx.bitrate_string)

    return ctx



from typing import List

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution


class EncoderRav1e(Encoder):
    def get_pretty_name(self) -> str:
        return "RAV1E"

    def get_encode_commands(self) -> List[str]:
        vec = [
            self.get_ffmpeg_pipe_command(),
            "|",
            get_binary("rav1e"),
            "-",
            f"-o {self.output_path}",
            f"-y",
            f"--threads {self.threads}",
            f"--keyint {self.keyint}",
            f"--speed {self.speed}",
            f"--photon-noise {self.grain_synth}",
        ]
        if self.tile_rows != -1:
            vec += [f"--tile-rows {self.tile_rows}"]
        if self.tile_cols != -1:
            vec += [f"--tile-cols {self.tile_cols}"]

        match self.rate_distribution:
            case EncoderRateDistribution.CQ:
                vec.append(f"--quantizer {self.crf * 4}")
            case EncoderRateDistribution.VBR:
                vec.append(f"--bitrate {self.bitrate}")
            case _:
                raise NotImplementedError

        if self.hdr:
            vec.append(f"--mastering-display {self.svt_master_display}")
            vec.append(
                f"--content-light {self.maximum_content_light_level},{self.maximum_frame_average_light_level}"
            )
            vec.append(f"--primaries {self.color_primaries}")
            vec.append(f"--transfer {self.transfer_characteristics}")
            vec.append(f"--matrix {self.matrix_coefficients}")

        return [" ".join(vec)]

    def get_chunk_file_extension(self) -> str:
        return ".ivf"

    def get_version(self) -> str:
        o = run_cli(get_binary("rav1e") + " --version").get_output()
        return o.split("\n")[0].replace("rav1e ", "")


    def get_codec(self) -> Codec:
        return Codec.av1



import os
import re

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.util.path import PathAlabama
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.impl.ssimu2 import Ssimu2Options
from alabamaEncode.metrics.impl.vmaf import VmafOptions
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.metrics.options import MetricOptions
from alabamaEncode.scene.chunk import ChunkObject


def calculate_metric(
    options=None,
    distorted_path: PathAlabama = None,
    reference_path: PathAlabama = None,
    chunk: ChunkObject = None,
    metric: Metric = Metric.VMAF,
):
    if chunk is None:
        _chunk = ChunkObject()
        if distorted_path is None or reference_path is None:
            raise ValueError(
                "distorted_path or reference_path cannot be None when chunk is None"
            )
        _chunk.chunk_path = distorted_path.get()
        _chunk.path = reference_path.get()
    else:
        _chunk = chunk

    match metric:
        case Metric.VMAF:
            from alabamaEncode.metrics.impl.vmaf import calc_vmaf

            return calc_vmaf(
                chunk=_chunk,
                vmaf_options=options if options is not None else VmafOptions(),
            )
        case Metric.SSIMULACRA2:
            from alabamaEncode.metrics.impl.ssimu2 import calc_ssimu2

            return calc_ssimu2(
                chunk=_chunk,
                ssimu2_options=options if options is not None else Ssimu2Options(),
            )
        case _:
            raise NotImplementedError(f"Metric {metric} not implemented")


def cleanup_input_pipes(output: dict):
    if os.path.exists(output["ref_pipe"]):
        os.remove(output["ref_pipe"])
    if os.path.exists(output["dist_pipe"]):
        os.remove(output["dist_pipe"])


def get_input_pipes(chunk: ChunkObject, options: MetricOptions) -> dict:
    """
    Create two named pipes that will output distorted and reference yuv frames,
    return the pipe paths and the commands that will feed them
    """

    video_filters = options.video_filters

    assert os.path.exists(chunk.path)
    assert os.path.exists(chunk.chunk_path)

    random_bit = os.urandom(16).hex()
    pipe_ref_path = f"/tmp/{os.path.basename(chunk.path)}_{random_bit}.pipe"
    pipe_dist_path = f"/tmp/{os.path.basename(chunk.chunk_path)}_{random_bit}.pipe"

    dist_filter = ""

    if options.ref is not None:
        comparison_scaling = f"scale={options.ref.__str__()}"

        vf = []
        if video_filters != "":
            for _filter in video_filters.split(","):
                if not re.match(r"scale=[0-9-]+:[0-9-]+", _filter):
                    vf.append(_filter)

        vf.append(comparison_scaling)
        video_filters = ",".join(vf)

        dist_filter = f" -vf {comparison_scaling} "

    if options.denoise_reference:
        # before scaling use the `vaguedenoiser` filter
        vf = video_filters.split(",")
        # add to the front of a list
        vf.insert(0, "vaguedenoiser")
        video_filters = ",".join(vf)

    video_filters = ",".join([f for f in video_filters.split(",") if f != ""])

    if video_filters != "":
        video_filters = f" -vf {video_filters} "

    ref_pipe_command = (
        f"{get_binary('ffmpeg')} -v error -nostdin -hwaccel auto {chunk.get_ss_ffmpeg_command_pair()}"
        f" -pix_fmt yuv420p10le -an -sn -strict -1 {video_filters} -f yuv4mpegpipe - > {pipe_ref_path}"
    )
    dist_pipe_command = (
        f'{get_binary("ffmpeg")} -v error -nostdin -filmgrain 0 -hwaccel auto -i "{chunk.chunk_path}" '
        f"-pix_fmt yuv420p10le -an -sn -strict -1 {dist_filter} -f yuv4mpegpipe - > {pipe_dist_path}"
    )

    # TODO: WINDOWS SUPPORT
    run_cli(f"mkfifo {pipe_ref_path}")
    run_cli(f"mkfifo {pipe_dist_path}")

    # check if both pipes are created
    assert os.path.exists(pipe_ref_path)
    assert os.path.exists(pipe_dist_path)

    return {
        "ref_pipe": pipe_ref_path,
        "dist_pipe": pipe_dist_path,
        "ref_command": ref_pipe_command,
        "dist_command": dist_pipe_command,
    }


def get_metric_from_stats(
    stats: EncodeStats,
    statistical_representation: str = "mean",
) -> float:
    match statistical_representation:
        case "mean":
            return stats.metric_results.mean
        case "harmonic_mean":
            return stats.metric_results.harmonic_mean
        case "max":
            return stats.metric_results.max
        case "min":
            return stats.metric_results.min
        case "median":
            return stats.metric_results.percentile_50
        case "percentile_1":
            return stats.metric_results.percentile_1
        case "percentile_5":
            return stats.metric_results.percentile_5
        case "percentile_10":
            return stats.metric_results.percentile_10
        case "percentile_25":
            return stats.metric_results.percentile_25
        case "percentile_50":
            return stats.metric_results.percentile_50
        case _:
            raise Exception(
                f"Unknown statistical_representation {statistical_representation}"
            )



import json
import os
import pickle

import keras
from keras.src.layers import Dense, LSTM, MaxPooling1D, Normalization
from keras_tuner import BayesianOptimization
from sklearn.model_selection import train_test_split
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map

from alabamaEncode.ai_vmaf.aom_firstpass import aom_extract_firstpass_data
from alabamaEncode.core.bin_utils import register_bin
from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.scene.chunk import ChunkObject
from alabamaEncode.scene.sequence import ChunkSequence
from alabamaEncode.scene.split import get_video_scene_list_skinny


def get_chunks_of_file(path, env, video_filters=""):
    scenes: ChunkSequence = get_video_scene_list_skinny(
        input_file=path, cache_file_path=f"{env}/sceneCache_{os.path.basename(path)}.pt"
    )

    for c in scenes.chunks:
        c.video_filters = video_filters

    return scenes.chunks


def get_chunks_for_analysis(env):
    analyzed_chunks = []

    # analyzed_chunks += get_chunks_of_file("test.mp4", "crop=etc", env=env)
    # ...

    # analyzed_chunks += get_chunks_of_file(
    #     "/mnt/data/objective-1-fast/Netflix_Aerial_1920x1080_60fps_8bit_420_60f.y4m",
    #     env=env,
    # )
    #
    # analyzed_chunks += get_chunks_of_file(
    #     "/mnt/data/objective-1-fast/Netflix_Crosswalk_1920x1080_60fps_8bit_420_60f.y4m",
    #     env=env,
    # )

    analyzed_chunks += get_chunks_of_file(
        "/home/kokoniara/coraline_sc.mkv",
        env=env,
        video_filters=f"scale=1920:-2,{Ffmpeg.get_tonemap_vf()}",
    )

    # analyzed_chunks += get_chunks_of_file(
    #     "/home/kokoniara/ep3_halo_test.mkv",
    #     env=env,
    #     video_filters=f"crop=3840:1920:0:120,scale=1920:-2,{Ffmpeg.get_tonemap_vf()}",
    # )

    analyzed_chunks += get_chunks_of_file(
        "/mnt/data/downloads/Halo.S01.2160p.UHD.BluRay.Remux.HDR.DV.HEVC.Atmos-PmP/Halo.S01E05.Reckoning.2160p.UHD"
        ".BluRay.Remux.HDR.DV.HEVC.Atmos-PmP.mkv",
        env=env,
        video_filters=f"crop=3840:1920:0:120,scale=1920:-2,{Ffmpeg.get_tonemap_vf()}",
    )

    analyzed_chunks += get_chunks_of_file("/mnt/data/liveaction_bright.mkv", env)[:5]

    analyzed_chunks += get_chunks_of_file(
        "/mnt/data/liveAction_normal.mp4",
        env=env,
        video_filters="crop=3808:1744:28:208,scale=1920:-2",
    )

    analyzed_chunks += get_chunks_of_file(
        "/home/kokoniara/howlscastle_test.mkv",
        env=env,
        video_filters="crop=1920:1024:0:28",
    )

    analyzed_chunks += get_chunks_of_file(
        "/home/kokoniara/ahs_s12_e05_cliptest.mp4", env=env
    )

    analyzed_chunks += get_chunks_of_file(
        "/mnt/data/liveAction_highMotion.mkv",
        env=env,
        video_filters="crop=1920:800:0:140",
    )

    analyzed_chunks += get_chunks_of_file(
        "/home/kokoniara/lesssons_in_methistry.mkv",
        env=env,
        video_filters=f"scale=1920:-2,{Ffmpeg.get_tonemap_vf()}",
    )

    for i, c in enumerate(analyzed_chunks):
        c.chunk_path = os.path.join(env, f"{i}.ivf")
        c.chunk_index = i

    return analyzed_chunks


def process_chunk(chunk):
    cache_file_name = chunk.chunk_path + ".cache"

    if os.path.exists(cache_file_name):
        with open(cache_file_name, "rb") as f:
            stats = pickle.load(f)
            tqdm.write(f"Found & Loaded cache stat for {chunk.chunk_path}")
            return stats
    else:
        tqdm.write(f"[{chunk.chunk_path}] Cache not found, ignoring for now")
        return None
        enc = EncoderSvt()
        # enc = EncoderX264()
        enc.chunk = chunk
        if hasattr(chunk, "video_filters") is False or chunk.video_filters is None:
            chunk.video_filters = ""
        enc.video_filters = chunk.video_filters

        enc.speed = 5
        enc.passes = 1
        enc.threads = 2
        enc.rate_distribution = EncoderRateDistribution.CQ

        local_stats = {}
        # for crf in range(18, 55, 2):
        for crf in [
            12,
            15,
            17,
            19,
            21,
            23,
            25,
            27,
            29,
            31,
            34,
            37,
            40,
            43,
            46,
            49,
            52,
            55,
            58,
        ]:
            enc.crf = crf
            enc.output_path = os.path.join(
                os.path.dirname(chunk.chunk_path),
                f"{crf}_{chunk.chunk_index}{enc.get_chunk_file_extension()}",
            )
            local_stats[str(crf)] = enc.run(calculate_vmaf=True).__dict__()
            if os.path.exists(enc.output_path):
                os.remove(enc.output_path)

        features = Ffmpeg.get_content_features(chunk, vf=chunk.video_filters)

        aom_features = aom_extract_firstpass_data(chunk, vf=chunk.video_filters)

        stats = {
            "features": features,
            "chunk_stats": local_stats,
            "aom_features": aom_features,
        }

        with open(cache_file_name, "wb") as f:
            pickle.dump(stats, f)
        if not os.path.exists(cache_file_name):
            raise Exception("Failed to save cache file")

    return stats


def process_chunks(env, chunks, max_workers=4):
    stats_file = os.path.join(env, "stats.pkl")
    stats = []

    if os.path.exists(stats_file):
        with open(stats_file, "rb") as f:
            return pickle.load(f)

    for item in process_map(
        process_chunk, chunks, max_workers=max_workers, chunksize=1
    ):
        if item is not None:
            stats.append(item)

    with open(stats_file, "wb") as f:
        pickle.dump(stats, f)

    return stats


def prepare_data(stats):
    features_list = []
    labels_list = []

    for chunk in stats:
        crf_values = list(
            chunk["chunk_stats"].keys()
        )  # Get the list of CRF values in the chunk

        features_chunk = []
        for frame in chunk["features"]:
            frame_features = [
                float(value) for value in chunk["features"][frame].values()
            ]
            features_chunk.append(frame_features)
        for crf in crf_values:
            vmaf_at_crf = chunk["chunk_stats"][crf]["vmaf"]
            print(f"VMAF at crf {crf}: {vmaf_at_crf}")
            features_with_vmaf = features_chunk.copy()
            for frame in features_with_vmaf:
                frame.append(float(vmaf_at_crf))
            features_list.append(features_with_vmaf)

            labels = [0] * len(crf_values)
            # 1 at the index of the crf value
            labels[crf_values.index(crf)] = 1
            labels_list.append(labels)

    return features_list, labels_list


def process_chunk_test():
    register_bin("SvtAv1EncApp", "/home/kokoniara/.local/opt/SvtAv1EncApp")
    test_chunk = ChunkObject(
        path="/mnt/data/objective-1-fast/Netflix_RollerCoaster_1280x720_60fps_8bit_420_60f.y4m"
    )
    env = os.path.abspath("./ai_feature_extract_test")
    if not os.path.exists(env):
        os.mkdir(env)

    test_chunk.chunk_path = os.path.join(env, "test.ivf")

    stats = process_chunk(test_chunk)
    print(stats)


def process_chunks_test():
    register_bin("SvtAv1EncApp", "/home/kokoniara/.local/opt/SvtAv1EncApp")
    env = os.path.abspath("./ai_feature_extract_test")
    if not os.path.exists(env):
        os.mkdir(env)

    chunks = get_chunks_for_analysis(env)
    stats = process_chunks(env, chunks)
    print(json.dumps(stats, indent=4))


def test_nn_impl():
    register_bin("SvtAv1EncApp", "/home/kokoniara/.local/opt/SvtAv1EncApp")
    env = os.path.abspath("./ai_feature_extract_test")
    if not os.path.exists(env):
        os.mkdir(env)

    # getting data
    chunks = get_chunks_for_analysis(env)
    print(f"Starting processing of {len(chunks)} chunks")
    stats = process_chunks(env, chunks)

    # preparing data
    features, labels = prepare_data(stats)

    print(f"Using {len(features)} chunks for training")

    # Turn into numpy arrays
    # features = np.array(features)
    # labels = np.array(labels)
    #
    # labels /= 100

    labels = [x / 100 for x in labels]

    # training model
    X_train, X_test, y_train, y_test = train_test_split(features, labels, test_size=0.2)

    normalization_layer = Normalization(
        input_shape=(len(features[0]), len(features[0][0])), name="frame_features"
    )

    normalization_layer.adapt(X_train)

    # Define the model building function for keras-tuner
    def build_model_lstm(hp):
        model = keras.Sequential()
        num_layers = hp.Int("num_layers", 1, 4)

        model.add(normalization_layer)
        for layer in range(num_layers):
            if layer < num_layers - 1:
                model.add(
                    LSTM(
                        units=hp.Int(
                            f"units_{layer}", min_value=32, max_value=480, step=32
                        ),
                        return_sequences=True,
                    )
                )
            else:
                model.add(
                    LSTM(
                        units=hp.Int(
                            "units_last", min_value=32, max_value=480, step=32
                        ),
                    )
                )

        model.add(Dense(1, activation="linear", name="predicted_crf"))
        model.compile(optimizer="adam", loss="mse")
        return model

    def build_model_dense(hp):
        model = keras.Sequential()
        model.add(
            MaxPooling1D(
                pool_size=hp.Int("pool_size", min_value=2, max_value=10, step=2),
                input_shape=(X_train.shape[1], X_train.shape[2]),
            )
        )
        for layer in range(hp.Int("num_layers", 1, 4)):
            model.add(
                Dense(
                    units=hp.Int("units", min_value=32, max_value=512, step=32),
                    activation="relu",
                )
            )

        model.add(Dense(1, activation="linear", name="predicted_crf"))

        model.compile(optimizer="adam", loss="mse")
        return model

    # Set up the tuner
    tuner = BayesianOptimization(
        build_model_lstm,
        # build_model_dense,
        objective="val_loss",
        max_trials=10,
    )

    # Perform the hyperparameter search
    # tuner.search(X_train, y_train, epochs=5, validation_data=(X_test, y_test))
    tuner.search(
        X_train,
        y_train,
        epochs=100,
        validation_data=(X_test, y_test),
        callbacks=[
            keras.callbacks.EarlyStopping(monitor="loss", patience=4),
            keras.callbacks.TensorBoard("/tmp/tb_logs"),
        ],
    )

    # Get the best model
    best_model = tuner.get_best_models()[0]

    # plot the learning curve of the best model

    # Evaluate the best model on the test set
    loss = best_model.evaluate(X_test, y_test)
    print(f"Test Loss: {loss}")

    # Do a prediction using the best model
    predictions = best_model.predict(X_test)
    print(f"Predicted: {predictions.flatten()}, Actual: {y_test}")

    # for all the chunks fo a prediction and print it
    for data, i in zip(features, range(len(features))):
        prediction = best_model.predict(data.reshape((1, *data.shape)))
        print(f"Chunk {i} predicted crf: {prediction.flatten()} real crf: {labels[i]}")

    # Save the model
    best_model.save(os.path.join(env, "model.keras"))


def test_nn_two():
    register_bin("SvtAv1EncApp", "/home/kokoniara/.local/opt/SvtAv1EncApp")
    env = os.path.abspath("./ai_feature_extract_test")
    if not os.path.exists(env):
        os.mkdir(env)

    # getting data
    chunks = get_chunks_for_analysis(env)
    print(f"Starting processing of {len(chunks)} chunks")
    stats = process_chunks(env, chunks)

    # preparing data
    features, labels = prepare_data(stats)

    # describe data using pd
    import pandas as pd

    df = pd.DataFrame(features)
    print(df)
    print("Features shape")
    print(df.shape)

    df = pd.DataFrame(labels)
    print("\nLabels Shape")
    print(df.shape)

    print(f"Using {len(features)} chunks for training")


if __name__ == "__main__":
    # process_chunk_test()
    # process_chunks_test()
    # test_nn_impl()
    test_nn_two()



"""
Testing enable-overlays.md in svtav1 using crf
"""
import os

from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.experiments.util.ExperimentUtil import (
    run_tests_across_range,
    read_report,
)

experiment_name = "Testing SvtAv1 Overlays --enable-overlays.md=1, speed 4"

test_env = os.getcwd() + "/data/overlays/"
if not os.path.exists(test_env):
    os.makedirs(test_env)

report_path = test_env + experiment_name + " CRF.json"


def analyze():
    df = read_report(report_path)

    print(df)

    vmaf1percenttile_change_arr = []
    size_change_arr = []
    vmaf_change_arr = []
    ssim_change_arr = []
    bpv_change_arr = []

    for key, group in df.groupby("basename"):
        print(key)
        control = 0
        enc1 = 0
        for key_j, group_j in group.groupby("test_group"):
            print("")
            # avg 1%tile vmaf
            score = group_j["vmaf_percentile_1"].mean()
            size_score = group_j["size"].mean()
            vmaf_score = group_j["vmaf"].mean()
            ssim_score = group_j["ssim"].mean()

            # how much bits per vmaf
            bpv_score = group_j["size"].mean() / group_j["vmaf"].mean()

            if key_j == "control":
                control = score
                size_control = size_score
                vmaf_control = vmaf_score
                ssim_control = ssim_score
                bpv_control = bpv_score
            else:
                enc1 = score
                size_enc1 = size_score
                vmaf_enc1 = vmaf_score
                ssim_enc1 = ssim_score
                bpv_enc1 = bpv_score

        # show pct of difference between control and enc1

        vmaf1percenttile_change_arr.append((enc1 - control) / control * 100)
        size_change_arr.append((size_enc1 - size_control) / size_control * 100)
        vmaf_change_arr.append((vmaf_enc1 - vmaf_control) / vmaf_control * 100)
        ssim_change_arr.append((ssim_enc1 - ssim_control) / ssim_control * 100)
        bpv_change_arr.append((bpv_enc1 - bpv_control) / bpv_control * 100)

    print(f"enable-overlays.md=1 tests")
    print(
        f"overall VMAF 1%tile avg (positive mean better): {sum(vmaf1percenttile_change_arr) / len(vmaf1percenttile_change_arr)}%"
    )
    print(
        f"overall size avg (positive mean worse): {sum(size_change_arr) / len(size_change_arr)}%"
    )
    print(
        f"overall VMAF avg (positive mean better): {sum(vmaf_change_arr) / len(vmaf_change_arr)}%"
    )
    print(
        f"overall SSIM avg (positive mean better): {sum(ssim_change_arr) / len(ssim_change_arr)}%"
    )
    print(
        f"overall Bits Per vmaf avg (positive mean worse): {sum(bpv_change_arr) / len(bpv_change_arr)}%"
    )

    # for key, group in df.groupby("basename"):
    #     print(key)
    #     fig, ax = plt.subplots()
    #     for key_j, group_j in group.groupby("test_group"):
    #         group_j.plot(
    #             ax=ax,
    #             kind="scatter",
    #             x="size",
    #             y="vmaf",
    #             label=key_j,
    #             color=("red" if key_j == "control" else "blue"),
    #         )
    #
    #     # 3d plot rate vmaf_percentile_1 size colored by test_group
    #     # fig = plt.figure()
    #     # ax = fig.add_subplot(111, projection="3d")
    #     # for key_j, group_j in group.groupby("test_group"):
    #     #     ax.scatter(
    #     #         group_j["rate"],
    #     #         group_j["vmaf_percentile_1"],
    #     #         group_j["size"],
    #     #         label=key_j,
    #     #         color=("red" if key_j == "control" else "blue"),
    #     #     )
    #     # ax.set_xlabel("rate")
    #     # ax.set_ylabel("vmaf_percentile_1")
    #     # ax.set_zlabel("size")
    #
    #     plt.title(key)
    #     plt.savefig(test_env + key + ".png")

    pass


def test():
    enc_control = EncoderSvt()
    enc_control.speed = 4
    enc_control.threads = 12
    enc_control.svt_tune = 1
    enc_control.svt_overlay = 0

    enc_test = EncoderSvt()
    enc_test.speed = 4
    enc_test.threads = 12
    enc_test.svt_tune = 1
    enc_test.svt_overlay = 1

    run_tests_across_range(
        [enc_control, enc_test], title=experiment_name, test_env=test_env, skip_vbr=True
    )


if __name__ == "__main__":
    # test()

    analyze()



import os
from copy import deepcopy

from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.experiments.util.ExperimentUtil import (
    run_tests_across_range,
    read_report,
)

experiment_name = "Testing SvtAv1 Adaptive QP level mode (--aq-mode) 1 vs 2, speed 4"

test_env = os.getcwd() + "/data/aq/"
if not os.path.exists(test_env):
    os.makedirs(test_env)

report_path = test_env + experiment_name + " CRF.json"


def test():
    control = EncoderSvt()
    control.speed = 4
    control.threads = 12
    control.svt_aq_mode = 2

    enc_test = deepcopy(control)
    control.svt_aq_mode = 1

    run_tests_across_range(
        [control, enc_test],
        title=experiment_name,
        test_env=test_env,
        skip_vbr=True,
    )


def analyze():
    df = read_report(report_path)

    print(df)

    enc1_vmaf1percenttile_change_arr = []
    enc1_size_change_arr = []
    enc1_vmaf_change_arr = []
    enc1_ssim_change_arr = []
    enc1_bpv_change_arr = []
    enc1_time_encoding_change_arr = []

    for key, group in df.groupby("basename"):
        print(key)
        control = 0
        enc1 = 0
        for key_j, group_j in group.groupby("test_group"):
            print(key_j)
            # avg 1%tile vmaf
            score = group_j["vmaf_percentile_1"].mean()
            size_score = group_j["size"].mean()
            vmaf_score = group_j["vmaf"].mean()
            ssim_score = group_j["ssim"].mean()
            time_encoding_score = group_j["time_encoding"].mean()

            # how much bits per vmaf
            bpv_score = group_j["size"].mean() / group_j["vmaf"].mean()

            if key_j == "control":
                control = score
                size_control = size_score
                vmaf_control = vmaf_score
                ssim_control = ssim_score
                bpv_control = bpv_score
                time_encoding_control = time_encoding_score
            elif key_j == "enc1":
                enc1 = score
                size_enc1 = size_score
                vmaf_enc1 = vmaf_score
                ssim_enc1 = ssim_score
                bpv_enc1 = bpv_score
                time_encoding_enc1 = time_encoding_score

        # show pct of difference between control and enc1

        enc1_vmaf1percenttile_change_arr.append((enc1 - control) / control * 100)
        enc1_size_change_arr.append((size_enc1 - size_control) / size_control * 100)
        enc1_vmaf_change_arr.append((vmaf_enc1 - vmaf_control) / vmaf_control * 100)
        enc1_ssim_change_arr.append((ssim_enc1 - ssim_control) / ssim_control * 100)
        enc1_bpv_change_arr.append((bpv_enc1 - bpv_control) / bpv_control * 100)
        enc1_time_encoding_change_arr.append(
            (time_encoding_enc1 - time_encoding_control) / time_encoding_control * 100
        )

    print("ENC1 (AQ 1)")
    print(
        f"overall VMAF 1%tile avg (positive mean better): {sum(enc1_vmaf1percenttile_change_arr) / len(enc1_vmaf1percenttile_change_arr)}%"
    )
    print(
        f"overall size avg (positive mean worse): {sum(enc1_size_change_arr) / len(enc1_size_change_arr)}%"
    )
    print(
        f"overall VMAF avg (positive mean better): {sum(enc1_vmaf_change_arr) / len(enc1_vmaf_change_arr)}%"
    )
    print(
        f"overall SSIM avg (positive mean better): {sum(enc1_ssim_change_arr) / len(enc1_ssim_change_arr)}%"
    )
    print(
        f"overall Bits Per vmaf avg (positive mean worse): {sum(enc1_bpv_change_arr) / len(enc1_bpv_change_arr)}%"
    )
    print(
        f"overall Time Encoding avg (positive mean worse): {sum(enc1_time_encoding_change_arr) / len(enc1_time_encoding_change_arr)}%"
    )


if __name__ == "__main__":
    # test()
    analyze()



from alabamaEncode.conent_analysis.chunk.chunk_analyse_step import (
    ChunkAnalyzePipelineItem,
)
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.scene.chunk import ChunkObject


class CrfIndexesMap(ChunkAnalyzePipelineItem):
    def run(self, ctx: AlabamaContext, chunk: ChunkObject, enc: Encoder) -> Encoder:
        crf_map = self.crf_map.split(",")

        if len(crf_map) < chunk.chunk_index:
            raise Exception(
                f"crf_map is too short, {len(crf_map)} < {chunk.chunk_index}"
            )

        enc.rate_distribution = EncoderRateDistribution.CQ
        enc.passes = 1
        try:
            enc.crf = int(crf_map[chunk.chunk_index])
        except ValueError:
            raise Exception(f"crf_map is not a list of ints: {self.crf_map}")
        enc.svt_tune = 0
        enc.svt_overlay = 0

        return enc

    def __init__(self, crf_map: str):
        self.crf_map = crf_map



import os

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

import numpy as np
from keras.preprocessing import sequence

from alabamaEncode.core.ffmpeg import Ffmpeg

import keras

model = None


def predict_crf(chunk, vf=""):
    global model
    if model is None:
        model = keras.models.load_model(
            "/home/kokoniara/dev/VideoSplit/alabamaEncode/ai_vmaf/"
            "train/ai_feature_extract_test/model.keras"
        )

    features_dict = Ffmpeg.get_content_features(chunk, vf=vf)

    # do this weird dance to get the data in the right shape

    features = []
    for frame in features_dict:
        frame_features = [float(value) for value in features_dict[frame].values()]
        features.append(frame_features)

    features = np.array([features])

    features = sequence.pad_sequences(
        features, maxlen=232, dtype="float32", padding="post"
    )

    features = features.reshape((1, *features.shape))

    crf = model.predict(features[0], verbose=0)[0][0]

    crf *= 100

    return int(crf)




import re

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.scene.chunk import ChunkObject


def get_video_psnr(distorted_path, in_chunk: ChunkObject = None):
    null_ = in_chunk.create_chunk_ffmpeg_pipe_command()
    null_ += f" | {get_binary('ffmpeg')} -hide_banner -i - "

    null_ += f" -i {distorted_path} -filter_complex psnr -f null -"

    result_string = run_cli(null_).get_output()
    try:
        # Regex to get avrage score out of:
        # [Parsed_psnr_0 @ 0x55f6705fa200] PSNR y:49.460782 u:52.207017 v:51.497660 average:50.118351
        # min:48.908778 max:51.443411

        psnr_score = float(re.search(r"average:([\d.]+)", result_string).group(1))
        return psnr_score
    except AttributeError:
        print(f"Failed getting psnr comparing {distorted_path} agains {in_chunk.path}")
        print(null_)
        return 0



from enum import Enum


class ComparisonDisplayResolution(Enum):
    """
    Most common viewer screen resolutions
    """

    HD = 1  # 720p
    FHD = 2  # 1080p
    UHD = 3  # 4k

    @staticmethod
    def from_string(resolution: str):
        if resolution == "HD":
            return ComparisonDisplayResolution.HD
        elif resolution == "FHD":
            return ComparisonDisplayResolution.FHD
        elif resolution == "UHD":
            return ComparisonDisplayResolution.UHD
        else:
            raise ValueError("Unknown display resolution")

    def __str__(self):
        if self == ComparisonDisplayResolution.HD:
            return "1280:720"
        elif self == ComparisonDisplayResolution.FHD:
            return "1920:1080"
        elif self == ComparisonDisplayResolution.UHD:
            return "3840:2160"
        else:
            raise ValueError("Unknown display resolution")



import asyncio
import json
import os
import socket
import time

import psutil
import requests
import websockets
from tqdm import tqdm


class WebsiteUpdate:

    def __init__(self, ctx):
        self.ctx = ctx
        self.current_step_callback = None
        self.proc_done_callback = None
        self.proc_done = 0
        self.current_step_name = "idle"
        self.last_update = None
        self.update_proc_throttle = 0
        self.update_max_freq_sec = 1
        self.ws_server = None

    async def update_website(self):
        api_url = os.environ.get("status_update_api_url", "")
        token = os.environ.get("status_update_api_token", "")
        if api_url != "":
            if token == "":
                print("Url is set, but token is not, not updating status api")
                return

            should_update_api = False

            if time.time() - self.update_proc_throttle > self.update_max_freq_sec:
                self.update_proc_throttle = time.time()
                should_update_api = True

            #         curl -X POST -d '{"action":"update","data":{"img":"https://domain.com/poster.avif","status":100,
            #         "title":"Show 2024 E01S01","phase":"Done"}}'
            #         -H 'Authorization: Bearer token' 'https://domain.com/update'

            status_data = {
                "action": "update",
                "data": {
                    "img": self.ctx.poster_url,
                    "status": round(self.proc_done, 1),  # rounded
                    "title": self.ctx.get_title(),
                    "phase": self.current_step_name,
                },
            }

            if self.last_update == status_data:
                return

            self.last_update = status_data

            if should_update_api:
                try:
                    requests.post(
                        api_url + "/statuses/update",
                        json=status_data,
                        headers={"Authorization": f"Bearer {token}"},
                    )
                except Exception as e:
                    self.ctx.log(f"Failed to update status api: {e}")

                self.ctx.log("Updated encode status api")

            #  curl -X POST -d '{"action":"update","data":{"id":"kokoniara-B550MH",
            #  "status":"working on title", "utilization":95}}' -H 'Authorization: Bearer token'
            #  'http://domain.com/workers/update'

            worker_data = {
                "action": "update",
                "data": {
                    "id": (socket.gethostname()),
                    "status": f"Working on {self.ctx.get_title()}",
                    "utilization": int(psutil.cpu_percent()),
                    "ws_ip": self.ws_server.ip if self.ws_server is not None else "",
                },
            }

            if should_update_api:
                try:
                    requests.post(
                        api_url + "/workers/update",
                        json=worker_data,
                        headers={"Authorization": f"Bearer {token}"},
                    )
                except Exception as e:
                    self.ctx.log(f"Failed to worker update status api: {e}")

                self.ctx.log("Updated worker status api")

    def update_current_step_name(self, step_name):
        self.current_step_name = step_name
        if self.current_step_callback is not None:
            self.current_step_callback(step_name)

        asyncio.create_task(self.update_website())

    def update_proc_done(self, proc_done):
        self.proc_done = proc_done
        if self.proc_done_callback is not None:
            self.proc_done_callback(proc_done)

        # update max every minute the proc done
        asyncio.create_task(self.update_website())

    async def constant_updates(self):
        if (
            os.environ.get("status_update_api_url", "") == ""
            or os.environ.get("ws_update", "false") == "false"
        ):
            print(
                f"Not starting constant updates, {os.environ.get('status_update_api_url', '')} {os.environ.get('ws_update', 'false')}"
            )

            return

        tqdm.write("Starting constant updates")
        self.ws_server = WebsocketServer()
        self.ws_server.run()
        while True:
            worker_data = {
                "type": "worker",
                "data": {
                    "id": (socket.gethostname()),
                    "status": f"Working on {self.ctx.get_title()}",
                    "utilization": int(psutil.cpu_percent()),
                    "ws_ip": self.ws_server.ip if self.ws_server is not None else "",
                },
            }
            status_data = {
                "type": "status",
                "data": {
                    "img": self.ctx.poster_url,
                    "status": round(self.proc_done, 1),  # rounded
                    "title": self.ctx.get_title(),
                    "phase": self.current_step_name,
                },
            }

            await self.ws_server.publish(json.dumps(worker_data), type="worker")
            await self.ws_server.publish(json.dumps(status_data), type="status")
            await asyncio.sleep(0.5)


class WebsocketServer:
    def __init__(self, port=6542):
        self.port = port
        self.clients = set()
        self.loop = asyncio.get_event_loop()
        self.server = None
        self.running = False
        # get local ip
        self.ip = socket.gethostbyname(socket.gethostname())
        self.last_worker_data = None
        self.last_status_data = None

    async def publish(self, data, type="worker"):
        if type == "worker":
            self.last_worker_data = data
        elif type == "status":
            self.last_status_data = data
        for client in self.clients:
            await client.send(data)

    async def handler(self, websocket):
        self.clients.add(websocket)
        # tqdm.write(f"New client connected, total clients: {len(self.clients)}")
        try:
            async for message in websocket:
                if message == "worker":
                    await websocket.send(self.last_worker_data)
                elif message == "status":
                    await websocket.send(self.last_status_data)
        finally:
            # print("Client disconnected")
            self.clients.remove(websocket)

    async def start(self):
        self.running = True
        self.server = await websockets.serve(
            self.handler, host="0.0.0.0", port=self.port
        )
        await self.server.wait_closed()

    def run(self):
        self.ws_task = self.loop.create_task(self.start())

    def stop(self):
        self.server.close()
        self.ws_task.cancel()
        self.running = False
        self.loop.close()



# get broken/backend url from env
import os
import traceback
from typing import Any

from celery import Celery

from alabamaEncode.parallel_execution.command import BaseCommandObject

BROKER_URL = os.getenv("BROKER_URL", "redis://" + os.getenv("REDIS_HOST", "localhost"))
BACKEND_URL = os.getenv(
    "BACKEND_URL", "redis://" + os.getenv("REDIS_HOST", "localhost")
)

app = Celery("ThaVaidioEncoda", broker=BROKER_URL, backend=BACKEND_URL)

app.conf.update(
    worker_concurrency=8,
    worker_autoscaler="alabamaEncode.CeleryAutoscaler.DAAutoscaler",
    task_serializer="pickle",
    result_serializer="pickle",
    accept_content=["pickle"],
    broker_connection_retry_on_startup=True,
)


@app.task(bind=True)
def run_command_on_celery(self, command: BaseCommandObject) -> Any:
    try:
        return command.run()
    except:
        traceback.print_exc()



import hashlib
import os


def parse_paths(ctx):
    """
    Sets up the paths
    """
    ctx.output_file = os.path.abspath(ctx.output_file)
    ctx.raw_input_file = os.path.abspath(ctx.raw_input_file)

    # turn tempfolder into a full path
    ctx.output_folder = os.path.normpath(os.path.dirname(ctx.output_file) + "/")

    # derive random part from the first few letters of the sha1 hash of title
    random_part = hashlib.sha1(ctx.get_title().encode("utf-8")).hexdigest()[:5]

    ctx.temp_folder = os.path.join(ctx.output_folder, f".alabamatemp-{random_part}")
    if not os.path.exists(ctx.temp_folder):
        os.makedirs(ctx.temp_folder)

    ctx.temp_folder += "/"

    ctx.input_file = os.path.join(ctx.temp_folder, "temp.mkv")

    if not os.path.exists(ctx.raw_input_file):
        print(f"Input file {ctx.raw_input_file} does not exist")
        quit()

    # symlink input file to temp folder
    if not os.path.exists(ctx.input_file):
        os.symlink(ctx.raw_input_file, ctx.input_file)
        if not os.path.exists(ctx.input_file):
            print(f"Failed to symlink input file to {ctx.input_file}")
            quit()

    return ctx



import os
import shutil

from alabamaEncode.core.util.timer import Timer
from alabamaEncode.encoder.impl.X264 import EncoderX264
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.scene.chunk import ChunkObject
from alabamaEncode.scene.concat import VideoConcatenator
from alabamaEncode.scene.scene_detection import scene_detect

counter = 0

if __name__ == "__main__":
    env = "./experiments/CUCKS6"
    input_file = "/mnt/data/60seccut_1500offset.mkv"

    env = os.path.abspath(env)

    # if os.path.exists(env):
    #     shutil.rmtree(env)

    if not os.path.exists(env):
        os.makedirs(env)

    enc = EncoderX264()
    # enc.chunk = ChunkObject(
    #     path="/home/kokoniara/showsEncode/HALO (2022)/s1/e5/lossless.mkv"
    # )
    enc.chunk = ChunkObject(
        path="/mnt/data/objective-1-fast/Netflix_PierSeaside_1920x1080_60fps_8bit_420_60f.y4m"
    )
    # enc.grain_synth = 3
    enc.speed = 1
    enc.rate_distribution = EncoderRateDistribution.CQ
    # enc.crf = 20
    # enc.qm_enabled = 1
    # enc.qm_min = 0
    # enc.qm_max = 15
    # enc.threads = 12
    # enc.output_path = os.path.devnull

    # enc.matrix_coefficients = "bt2020-ncl"
    # enc.color_primaries = "bt2020"
    # enc.transfer_characteristics = "smpte2084"
    # enc.chroma_sample_position = "topleft"
    # enc.maximum_content_light_level = 358
    # enc.maximum_frame_average_light_level = 92
    # enc.svt_master_display = (
    #     "G(0.265,0.69)B(0.15,0.06)R(0.68,0.32)WP(0.3127,0.329)L(1000.0,0.005)"
    # )
    # enc.passes = 1
    # enc.hdr = True
    # enc.svt_cli_path = "/home/kokoniara/.local/opt/SvtAv1EncApp"

    # enc.video_filters = "atadenoise,hqdn3d=4,unsharp=7:7:0.5"
    enc.video_filters = "scale=1920:-2"

    # context = AlabamaContext()
    # context.vmaf = 95
    # context.log_level = 2
    # context.temp_folder = env
    # context.crf_model_weights = "0,0,0,0,1"

    # opt = TargetVmaf(alg_type="binary", probes=4, probe_speed=4)
    # enc = opt.run(context, enc.chunk, enc)
    #
    # print(enc.crf)
    # enc.output_path = os.path.join(env, "output.ivf")
    # print(enc.run(calculate_vmaf=True))

    # data = []

    scene_list = scene_detect(
        input_file=input_file,
        cache_file_path=env + "/sceneCache.pt",
        max_scene_length=10,
    )

    target_vmaf = 93
    epsilon_down = 0.3

    for chunk in scene_list.chunks:
        enc.chunk = chunk
        runs = []
        chunk_probes = os.path.join(env, f"{chunk.chunk_index}_probes")
        if not os.path.exists(chunk_probes):
            os.makedirs(chunk_probes)
        # for crf in range(10, 35):
        #     enc.crf = crf
        #     enc.output_path = os.path.join(env, f"{chunk_probes}/{crf}.mkv")
        #     stats: EncodeStats = enc.run(calculate_vmaf=True)
        #     stats.version = crf
        #     stats.basename = enc.output_path
        #     if stats.vmaf < target_vmaf - epsilon_down:
        #         break
        #     print(stats)
        #     cums.append(stats)

        # binary search version
        low_crf = 10
        high_crf = 35
        timer = Timer()
        timer.start("binary search")
        while low_crf <= high_crf:
            mid_crf = (low_crf + high_crf) // 2
            enc.crf = mid_crf
            enc.output_path = os.path.join(env, f"{chunk_probes}/{mid_crf}.mkv")
            timer.start(f"crf_{mid_crf}")
            stats: EncodeStats = enc.run(metric_to_calculate=Metric.VMAF)
            timer.stop(f"crf_{mid_crf}")
            stats.version = mid_crf
            stats.basename = enc.output_path
            runs.append(stats)
            print(stats)

            if abs(stats.vmaf - target_vmaf) <= epsilon_down:
                break
            elif stats.vmaf > target_vmaf:
                low_crf = mid_crf + 1
            else:
                high_crf = mid_crf - 1
        timer.stop("binary search")
        timer.finish(loud=True)

        closes_to_target = min(runs, key=lambda x: abs(x.vmaf - target_vmaf))
        # move closes to target to ../{index}.mkv
        shutil.move(
            closes_to_target.basename, os.path.join(env, f"{chunk.chunk_index}.mkv")
        )

    concat = VideoConcatenator(
        output=os.path.join(env, "output.mkv"), file_with_audio=input_file
    )
    concat.find_files_in_dir(
        folder_path=env,
        extension=".mkv",
    )
    concat.concat_videos()

    quit()

    # for crf in range(15, 40):
    #     enc.crf = crf
    #     enc.output_path = os.path.join(env, f"output_{crf}.mkv")
    #     stats: EncodeStats = enc.run(calculate_vmaf=True)
    #     stats.version = crf
    #     print(stats)
    #     dict_stats = stats.get_dict()
    #     data.append(dict_stats)
    #
    # json.dump(data, open(os.path.join(env, "data.json"), "w"))



import copy
import os
import subprocess

import cv2
import numpy as np

from alabamaEncode.conent_analysis.chunk.chunk_analyse_step import (
    ChunkAnalyzePipelineItem,
)
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.get_yuv_stream import get_yuv_frame_stream
from alabamaEncode.core.util.timer import Timer
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.scene.chunk import ChunkObject
from alabamaEncode.scene.scene_detection import scene_detect


class NewGrainSynth(ChunkAnalyzePipelineItem):
    def run(self, ctx: AlabamaContext, chunk: ChunkObject, enc: Encoder) -> Encoder:
        grain_synth_result = ctx.get_kv().get("grain_synth", chunk.chunk_path)

        if grain_synth_result is None:
            grain_synth_result = calc_grainsynth_of_scene_fast(
                chunk,
                scale_vf=ctx.scale_string,
                crop_vf=ctx.crop_string,
            )
            grain_synth_result = min(grain_synth_result, 18)

            enc.grain_synth = int(grain_synth_result)
            ctx.get_kv().set("grain_synth", chunk.chunk_path, grain_synth_result)
        else:
            enc.grain_synth = int(grain_synth_result)

        ctx.log(f"{chunk.log_prefix()}computed gs {enc.grain_synth}", category="grain")

        return enc


def measure_output_size(command):
    process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    total_size = 0

    while process.poll() is None:
        chunk = process.stdout.read(1024 * 1024)
        if chunk:
            total_size += len(chunk)

    # Ensure the process has finished
    process.wait()

    return total_size


def get_single_frame(chunk: ChunkObject, vf: str = "") -> np.ndarray:
    chunk: ChunkObject = copy.deepcopy(chunk)
    chunk.last_frame_index = chunk.first_frame_index + 1

    class TopTenHacksBigPythonDosentWantYouToKnow(Exception):
        pass

    def cb(yuv_frame):
        frame = np.frombuffer(yuv_frame.buffer, dtype=np.uint8)
        frame = frame.reshape((yuv_frame.headers["H"] * 3 // 2, yuv_frame.headers["W"]))
        frame = cv2.cvtColor(frame, cv2.COLOR_YUV2BGR_I420)
        raise TopTenHacksBigPythonDosentWantYouToKnow(frame)

    try:
        get_yuv_frame_stream(
            chunk,
            frame_callback=cb,
            vf=vf,
        )
    except TopTenHacksBigPythonDosentWantYouToKnow as ex:
        return ex.args[0]


def inference(data):
    from itertools import combinations_with_replacement

    n_samples, n_features = data.shape
    combinations = list(combinations_with_replacement(range(n_features), 2))
    X_poly = np.ones((n_samples, len(combinations)))

    for i, index_comb in enumerate(combinations):
        X_poly[:, i] = np.prod(data[:, index_comb], axis=1)

    new_data_poly = X_poly

    new_data_poly_b = np.c_[np.ones((new_data_poly.shape[0], 1)), new_data_poly]
    weights = np.array([34.31016125, 2333.23498342, -4837.40104381, 2474.86714525])
    res = np.dot(new_data_poly_b, weights)
    return int(res[0])


def calc_grainsynth_of_scene_fast(
    chunk: ChunkObject, scale_vf="", crop_vf="", return_source_values=False
) -> int:
    filter_vec = []
    if crop_vf != "" and crop_vf is not None:
        filter_vec.append(f"crop={crop_vf}")

    if scale_vf != "" and scale_vf is not None:
        filter_vec.append(f"scale={scale_vf}:flags=lanczos")

    denoise_weak = "vaguedenoiser"
    denoise_strong = "vaguedenoiser=threshold=6"

    timer = Timer()

    timer.start("scene_gs_approximation")

    common = (
        f"{get_binary('ffmpeg')} -v error "
        f"-y {chunk.get_ss_ffmpeg_command_pair()} -pix_fmt yuv420p10le -c:v huffyuv -an -f nut "
    )
    a = ",".join(filter_vec)
    if a != "":
        a = f" -vf {a}"

    timer.start("ref")
    ref_size = measure_output_size(f"{common} {a} -")
    timer.stop("ref")

    timer.start("strong")
    strong_size = measure_output_size(
        f"{common} -vf {",".join(filter_vec + [denoise_strong])} -"
    )
    timer.stop("strong")

    timer.start("weak")
    weak_size = measure_output_size(
        f"{common} -vf {",".join(filter_vec + [denoise_weak])} -"
    )
    timer.stop("weak")

    normalised_weak = weak_size / ref_size
    normalised_strong = strong_size / ref_size

    final_grain = inference(np.array([[normalised_weak, normalised_strong]]))

    timer.stop("scene_gs_approximation")
    timer.finish(loud=True)
    if return_source_values:
        return {
            "ref_size": ref_size,
            "weak_size": weak_size,
            "strong_size": strong_size,
        }
    else:
        return final_grain


def test():
    test_env = "./experiments/grain_synth/"
    test_env = os.path.abspath(test_env)

    if not os.path.exists(test_env):
        os.makedirs(test_env)

    scene_list = scene_detect(
        input_file="/home/kokoniara/ep3_halo_test.mkv",
        cache_file_path=test_env + "sceneCache.pt",
    )

    for chunk in scene_list.chunks:
        print(
            "calculated gs: "
            + str(
                calc_grainsynth_of_scene_fast(
                    chunk, crop_vf="3840:1920:0:120", scale_vf="1920:-2"
                )
            )
        )
        print("\n\n")


def test1():
    test_env = "./experiments/grain_synth/"
    test_env = os.path.abspath(test_env)

    if not os.path.exists(test_env):
        os.makedirs(test_env)

    scene_list = scene_detect(
        input_file="/home/kokoniara/ep3_halo_test.mkv",
        cache_file_path=test_env + "sceneCache.pt",
    )

    for chunk in scene_list.chunks:
        image_data = get_single_frame(
            chunk, vf="'crop=270:270:(in_w-270)/2:(in_h-270)/2'"
        )  # np.ndarray
        print(chunk.chunk_index)
        # save image to test_env
        cv2.imwrite(test_env + f"test{chunk.chunk_index}.png", image_data)

        filtered_image = cv2.Laplacian(image_data, cv2.CV_64F)

        # save the filtered image
        cv2.imwrite(
            test_env + f"filtered{chunk.chunk_index}.png",
            filtered_image,
        )

        high_frequency_energy = np.mean(np.abs(filtered_image))

        print(high_frequency_energy)


if __name__ == "__main__":
    # test()
    test1()



from typing import List

from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.impl.Aomenc import EncoderAom
from alabamaEncode.encoder.impl.Nvenc import EncoderNVENCH264
from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.encoder.impl.VaapiH264 import EncoderVaapiH264
from alabamaEncode.encoder.impl.VaapiH265 import EncoderVaapiH265
from alabamaEncode.encoder.impl.VappiAv1 import EncoderVaapiAV1
from alabamaEncode.encoder.impl.VideoToolbox import EncoderAppleHEVC
from alabamaEncode.encoder.impl.X264 import EncoderX264
from alabamaEncode.encoder.impl.X265 import EncoderX265
from alabamaEncode.encoder.impl.rav1e import EncoderRav1e
from alabamaEncode.encoder.impl.vp9 import EncoderVPX

encoders = [
    EncoderX265(),
    EncoderSvt(),
    EncoderAom(),
    EncoderX264(),
    EncoderVPX("vp9"),
    EncoderVPX("vp8"),
    EncoderVaapiH265(),
    EncoderVaapiH264(),
    EncoderVaapiAV1(),
    EncoderRav1e(),
    EncoderNVENCH264(),
    EncoderAppleHEVC(),
]


def get_all_encoder_strings() -> List[str]:
    return [enc.get_pretty_name() for enc in encoders]


def get_encoder_from_string(string: str) -> Encoder:
    # special case for vpx since one class represents two encoders
    if string == "VPX_VP9":
        return EncoderVPX("vp9")

    if string == "VPX_VP8":
        return EncoderVPX("vp8")

    for enc in encoders:
        if string == enc.get_pretty_name():
            enc_class = enc.__class__
            return enc_class()

    raise Exception(f'Could not find a encoder "{string}"')


if __name__ == "__main__":
    print(get_all_encoder_strings())
    print(get_encoder_from_string("VPX_VP9"))
    print(get_encoder_from_string("X264"))



import argparse

from argparse_range import range_action

from alabamaEncode.encoder.encoder_factory import (
    get_all_encoder_strings,
    get_encoder_from_string,
)
from alabamaEncode.encoder.impl.Svtenc import EncoderSvt


def read_args(ctx):
    """
    Parse the arguments for the program
    """
    parser = argparse.ArgumentParser(
        prog="AlabamaEncoder",
        description="AlabamaEncoder, one and only encoder framework for all your needs*",
        epilog="*not really",
    )

    subparsers = parser.add_subparsers(dest="command")

    worker_parser = subparsers.add_parser("worker", help="Celery worker mode")
    worker_parser.add_argument(
        "--workers",
        "-workers",
        type=int,
        help="Amount of multiprocess workers",
        default=2,
        dest="workers",
    )

    subparsers.add_parser("clear", help="clear celery queue")

    auto_thumbnailer = subparsers.add_parser(
        "autothumbnailer", help="Pick and extract perfect thumbnail frames"
    )

    auto_thumbnailer.add_argument("input", type=str, help="Input video file")

    auto_thumbnailer.add_argument(
        "--detect_faces",
        action="store_true",
        help="Detect faces in video",
        dest="detect_faces",
    )

    encode = subparsers.add_parser("encode", help="Typical encode mode")

    encode.add_argument("input", type=str, help="Input video file")
    encode.add_argument("output", type=str, help="Output video file")

    encode.add_argument(
        "--dont_encode_audio",
        help="Mux audio",
        action="store_false",
        dest="encode_audio",
    )

    encode.add_argument(
        "--audio_params",
        help="Audio params",
        type=str,
        default=ctx.audio_params,
        dest="audio_params",
    )

    encode.add_argument(
        "--celery",
        help="Encode on a celery cluster, that is at localhost",
        action="store_true",
        default=ctx.use_celery,
        dest="use_celery",
    )

    encode.add_argument(
        "--autocrop",
        help="Automatically crop the video",
        action="store_true",
        default=ctx.auto_crop,
        dest="auto_crop",
    )

    encode.add_argument(
        "--video_filters",
        type=str,
        default=ctx.prototype_encoder.video_filters,
        help="Override the crop, put your vf ffmpeg there, example "
        "scale=-2:1080:flags=lanczos,zscale=t=linear..."
        " make sure ffmpeg on all workers has support for the filters you use",
    )

    encode.add_argument(
        "--bitrate",
        help="Bitrate to use, `auto` for auto bitrate selection",
        type=str,
        default=str(ctx.prototype_encoder.bitrate),
    )

    encode.add_argument(
        "--overshoot",
        help="How much proc the vbr_perchunk_optimisation is allowed to overshoot",
        type=int,
        default=ctx.bitrate_overshoot,
        dest="bitrate_overshoot",
    )

    encode.add_argument(
        "--undershoot",
        help="How much proc the vbr_perchunk_optimisation is allowed to undershoot",
        type=int,
        default=ctx.bitrate_undershoot,
        dest="bitrate_undershoot",
    )

    encode.add_argument(
        "--vbr_perchunk_optimisation",
        help="Enable automatic bitrate optimisation per chunk",
        action="store_true",
        dest="vbr_perchunk_optimisation",
    )

    encode.add_argument(
        "--multiprocess_workers",
        "--workers",
        "-j",
        help="Number of workers to use for multiprocessing, if -1 the program will auto scale",
        type=int,
        default=ctx.multiprocess_workers,
        dest="multiprocess_workers",
    )

    encode.add_argument(
        "--ssim-db-target",
        type=float,
        default=ctx.ssim_db_target,
        help="What ssim dB to target when using auto bitrate, not recommended to set manually, "
        "otherwise 21.2 is a good starting point",
        dest="ssim_db_target",
    )

    encode.add_argument(
        "--crf",
        help="What crf to use",
        type=int,
        default=ctx.prototype_encoder.crf,
        action=range_action(0, 255),
        dest="crf",
    )

    encode.add_argument(
        "--encoder",
        help="What encoder to use",
        type=str,
        default=EncoderSvt().get_pretty_name(),
        choices=get_all_encoder_strings(),
        dest="encoder",
    )

    encode.add_argument(
        "--grain",
        help="Manually give the grainsynth value, 0 to disable, -1 for auto, -2 for auto per scene, -3 for new fast "
        "per scene grain",
        type=int,
        default=ctx.prototype_encoder.grain_synth,
        dest="grain",
    )

    encode.add_argument(
        "--vmaf_target",
        help="What vmaf to target when using bitrate auto",
        default=ctx.vmaf,
        type=int,
        dest="vmaf_target",
    )

    encode.add_argument(
        "--max_scene_length",
        help="If a scene is longer then this, it will recursively cut in the"
        " middle it to get until each chunk is within the max",
        type=int,
        default=ctx.max_scene_length,
        dest="max_scene_length",
    )

    scene_split_method_group = encode.add_mutually_exclusive_group()

    scene_split_method_group.add_argument(
        "--statically_sized_scenes",
        help="Instead of preforming scene detection do statically sized scenes at about 30secs",
        action="store_true",
        dest="statically_sized_scenes",
    )

    scene_split_method_group.add_argument(
        "--scene_merge",
        help="Merge scenes until they met the max scene length",
        action="store_true",
        dest="scene_merge",
    )

    encode.add_argument(
        "--no_crf_based_vmaf_targeting",
        "--crf_mode",
        help="per chunk, find a crf that hits target quality and encode using that",
        action="store_false",
        dest="crf_based_vmaf_targeting",
    )

    encode.add_argument(
        "--chunk_order",
        help="Encode chunks in a specific order",
        type=str,
        default=ctx.chunk_order,
        choices=[
            "random",
            "sequential",
            "length_desc",
            "length_asc",
            "sequential_reverse",
            "even",
        ],
        dest="chunk_order",
    )

    encode.add_argument(
        "--start_offset",
        help="Offset from the beginning of the video (in seconds), useful for cutting intros etc",
        default=ctx.start_offset,
        type=int,
        dest="start_offset",
    )

    encode.add_argument(
        "--end_offset",
        help="Offset from the end of the video (in seconds), useful for cutting end credits outtros etc",
        default=ctx.end_offset,
        type=int,
        dest="end_offset",
    )

    encode.add_argument(
        "--bitrate_adjust_mode",
        help="do a complexity analysis on each chunk individually and adjust "
        "bitrate based on that, can overshoot/undershoot a lot, "
        "otherwise do complexity analysis on all chunks ahead of time"
        " and budget it to hit target by normalizing the bitrate",
        type=str,
        default=ctx.bitrate_adjust_mode,
        choices=["chunk", "global"],
        dest="bitrate_adjust_mode",
    )

    encode.add_argument(
        "--log_level",
        help="Set the log level, 0 silent, 1 verbose",
        type=int,
        default=ctx.log_level,
        dest="log_level",
    )

    encode.add_argument(
        "--generate_previews",
        "-previews",
        help="Dont generate previews for encoded file",
        action="store_true",
        dest="generate_previews",
    )

    encode.add_argument(
        "--generate_stats",
        "-stats",
        help="Generate stats for encoded file",
        action="store_true",
        dest="generate_stats",
    )

    encode.add_argument(
        "--override_bad_wrong_cache_path",
        help="Override the check for input file path matching in scene cache loading",
        action="store_true",
        default=ctx.override_scenecache_path_check,
        dest="override_bad_wrong_cache_path",
    )

    encode.add_argument(
        "--hdr",
        "-hdr",
        help="Encode in HDR, if not specified and input is hdr it will automatically tonemap",
        action="store_true",
        default=ctx.prototype_encoder.hdr,
        dest="hdr",
    )

    encode.add_argument(
        "--crop_string",
        help="Crop string to use, eg `1920:1080:0:0`, `3840:1600:0:280`. Obtained using the `cropdetect` ffmpeg filter",
        type=str,
        default=ctx.crop_string,
        dest="crop_string",
    )

    encode.add_argument(
        "--scale_string",
        help="Scale string to use, eg. `1920:1080`, `1280:-2`, `1920:1080:force_original_aspect_ratio=decrease`",
        type=str,
        default=ctx.scale_string,
        dest="scale_string",
    )

    encode.add_argument(
        "--dry_run",
        help="Do not encode, just print what would be done",
        action="store_true",
        dest="dry_run",
    )

    encode.add_argument(
        "--title", help="Title of the video", type=str, default=ctx.title, dest="title"
    )

    encode.add_argument(
        "--encoder_flag_override",
        type=str,
        default=ctx.prototype_encoder.override_flags,
        help="Override the encoder flags with this string, write all params except paths",
        dest="encoder_flag_override",
    )

    encode.add_argument(
        "--encoder_speed_override",
        "--enc_speed",
        type=int,
        action=range_action(0, 10),
        default=ctx.prototype_encoder.speed,
        help="Override the encoder speed parameter",
        dest="encoder_speed_override",
    )

    encode.add_argument(
        "--crf_map",
        type=str,
        help="Map of crf <-> chunk index, for debugging purposes only",
        default=ctx.crf_map,
        dest="crf_map",
    )

    encode.add_argument(
        "--color-primaries",
        type=str,
        default=ctx.prototype_encoder.color_primaries,
        help="Color primaries",
        dest="color_primaries",
    )
    encode.add_argument(
        "--transfer-characteristics",
        type=str,
        default=ctx.prototype_encoder.transfer_characteristics,
        help="Transfer characteristics",
        dest="transfer_characteristics",
    )
    encode.add_argument(
        "--matrix-coefficients",
        type=str,
        default=ctx.prototype_encoder.matrix_coefficients,
        help="Matrix coefficients",
        dest="matrix_coefficients",
    )
    encode.add_argument(
        "--maximum_content_light_level",
        type=str,
        default=ctx.prototype_encoder.maximum_content_light_level,
        help="Maximum content light level",
        dest="maximum_content_light_level",
    )

    encode.add_argument(
        "--frame-average-light",
        type=str,
        default=ctx.prototype_encoder.maximum_frame_average_light_level,
        help="Maximum frame average light level",
        dest="frame_average_light",
    )
    encode.add_argument(
        "--chroma-sample-position",
        type=str,
        default=ctx.prototype_encoder.chroma_sample_position,
        help="Chroma sample position",
    )

    encode.add_argument(
        "--sub_file",
        type=str,
        default=ctx.sub_file,
        help="Subtitles file, eg .srt or .vvt",
        dest="sub_file",
    )

    encode.add_argument(
        "--crf_limits",
        type=str,
        help="limits for crf when targeting vmaf, comma separated eg 20,35",
        dest="crf_limits",
    )

    encode.add_argument(
        "--vmaf_phone_model",
        action="store_true",
        help="use vmaf phone model for auto crf tuning",
        default=ctx.vmaf_phone_model,
    )

    encode.add_argument(
        "--vmaf_4k_model",
        action="store_true",
        help="use vmaf 4k model for auto crf tuning",
        dest="vmaf_4k_model",
    )

    encode.add_argument(
        "--vmaf_subsample",
        type=int,
        help="compute scores only every N frames",
        dest="vmaf_subsample",
    )

    encode.add_argument(
        "--vmaf_no_motion",
        action="store_true",
        help="use vmaf no motion model for auto crf tuning",
        dest="vmaf_no_motion",
    )

    encode.add_argument(
        "--auto_accept_autocrop",
        action="store_true",
        help="Automatically accept autocrop",
        dest="auto_accept_autocrop",
    )

    encode.add_argument(
        "--resolution_preset",
        "-res",
        type=str,
        default=ctx.resolution_preset,
        help="Preset for the scale filter, possible choices are 4k 1440p 1080p 768p 720p 540p 480p 360p",
        dest="resolution_preset",
    )

    encode.add_argument(
        "--probe_count",
        type=int,
        default=ctx.probe_count,
        help="Max number of probes for metric targeting, higher is more accurate but slower",
        action=range_action(1, 10),
        dest="probe_count",
    )

    encode.add_argument(
        "--vmaf_probe_speed",
        dest="vmaf_probe_speed",
        type=int,
        default=ctx.vmaf_probe_speed,
        help="Override the speed for target vmaf probes",
        action=range_action(0, 10),
    )

    encode.add_argument(
        "--vmaf_reference_display",
        type=str,
        default=ctx.vmaf_reference_display,
        choices=["HD", "FHD", "UHD"],
        help="HD FHD UHD",
        dest="vmaf_reference_display",
    )

    encode.add_argument(
        "--vmaf_target_repesentation",
        type=str,
        default=ctx.vmaf_target_representation,
        help="vmaf target representation, default is mean",
        choices=[
            "mean",
            "min",
            "max",
            "harmonic_mean",
            "percentile_1",
            "percentile_5",
            "percentile_10",
            "percentile_25",
            "percentile_50",
        ],
        dest="vmaf_target_repesentation",
    )

    encode.add_argument(
        "--simple_denoise",
        action="store_true",
        help="use atadenoise on input, useful for x26 encoding with very noisy imputs and target vmaf, "
        "to be automated in the future",
        dest="simple_denoise",
    )

    encode.add_argument(
        "--dont_pin_to_cores",
        action="store_false",
        help="pin each chunk to a core",
        dest="dont_pin_to_cores",
    )

    encode.add_argument(
        "--niceness",
        type=int,
        default=ctx.prototype_encoder.niceness,
        help="nice the encoder process",
        dest="niceness",
    )

    encode.add_argument(
        "--print_analysis_logs",
        action="store_true",
        help="Print content analysis logs into console, like what crf did vmaf target pick etc",
        dest="print_analysis_logs",
    )

    encode.add_argument(
        "--poster_url",
        type=str,
        help="Url of poster for website updates",
        dest="poster_url",
    )

    encode.add_argument(
        "--offload_server",
        type=str,
        default="",
        help="if filled with a server address, will try to send a serialised job to that server",
        dest="offload_server",
    )

    encode.add_argument(
        "--metric_to_target",
        default=ctx.metric_to_target,
        choices=["vmaf", "ssimu2"],
        help="Uses all the vmaf target logic but a different metric",
        dest="metric_to_target",
    )

    encode.add_argument(
        "--dynamic_vmaf_target",
        action="store_true",
        help="Target vmaf and weight it against the bitrate, useful for lossy sources that trick vmaf into low scores",
        dest="dynamic_vmaf_target",
    )

    encode.add_argument(
        "--dynamic_vmaf_target_vbr",
        action="store_true",
        help="vmaf targeting but instead of tuning crf, we tune the bitrate and use variable bitrate encoding",
        dest="dynamic_vmaf_target_vbr",
    )

    encode.add_argument(
        "--tune",
        default=ctx.args_tune,
        type=str,
        choices=["fidelity", "appeal", "balanced"],
        help="Tune the encoder setting for a specific use case",
        dest="tune",
    )

    encode.add_argument(
        "--denoise_vmaf_ref",
        action="store_true",
        help="Denoise the vmaf reference",
        dest="denoise_vmaf_ref",
    )

    encode.add_argument(
        "--calc_final_vmaf",
        default=ctx.calc_final_vmaf,
        action="store_true",
        help="Calculate vmaf of the final chunk at the end",
        dest="calc_final_vmaf",
    )

    encode.add_argument(
        "--multi_res_pipeline",
        action="store_true",
        help="Create a optimised multi bitrate tier stream",
        dest="multi_res_pipeline",
    )

    encode.add_argument(
        "--throughput_scaling",
        action="store_true",
        help="Scale the multi-process workers based on throughput",
        dest="throughput_scaling",
    )

    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
        quit()

    if (
        args.command == "autothumbnailer"
        or args.command == "worker"
        or args.command == "clear"
    ):
        return ctx, args

    ctx.output_file = args.output
    ctx.output_folder = ctx.output_file
    ctx.raw_input_file = args.input
    ctx.prototype_encoder = get_encoder_from_string(args.encoder)
    ctx.prototype_encoder.grain_synth = args.grain
    ctx.log_level = args.log_level
    ctx.dry_run = args.dry_run
    ctx.ssim_db_target = args.ssim_db_target
    ctx.simple_denoise = args.simple_denoise
    ctx.vmaf = args.vmaf_target
    ctx.vbr_perchunk_optimisation = args.vbr_perchunk_optimisation
    ctx.crf_based_vmaf_targeting = args.crf_based_vmaf_targeting
    ctx.use_celery = args.use_celery
    ctx.prototype_encoder.override_flags = args.encoder_flag_override
    ctx.prototype_encoder.speed = args.encoder_speed_override
    ctx.multiprocess_workers = args.multiprocess_workers
    ctx.bitrate_adjust_mode = args.bitrate_adjust_mode
    ctx.bitrate_undershoot = args.bitrate_undershoot
    ctx.bitrate_overshoot = args.bitrate_overshoot
    ctx.prototype_encoder.crf = args.crf
    ctx.prototype_encoder.hdr = args.hdr
    ctx.max_scene_length = args.max_scene_length
    ctx.start_offset = args.start_offset
    ctx.end_offset = args.end_offset
    ctx.override_scenecache_path_check = args.override_bad_wrong_cache_path
    ctx.crop_string = args.crop_string
    ctx.scale_string = args.scale_string
    ctx.title = args.title
    ctx.chunk_order = args.chunk_order
    ctx.audio_params = args.audio_params
    ctx.generate_previews = args.generate_previews
    ctx.generate_stats = args.generate_stats
    ctx.encode_audio = args.encode_audio
    ctx.sub_file = args.sub_file
    ctx.prototype_encoder.color_primaries = args.color_primaries
    ctx.prototype_encoder.transfer_characteristics = args.transfer_characteristics
    ctx.prototype_encoder.matrix_coefficients = args.matrix_coefficients
    ctx.prototype_encoder.maximum_content_light_level = args.maximum_content_light_level
    ctx.prototype_encoder.maximum_frame_average_light_level = args.frame_average_light
    ctx.prototype_encoder.chroma_sample_position = args.chroma_sample_position
    ctx.prototype_encoder.video_filters = args.video_filters
    ctx.auto_crop = args.auto_crop
    ctx.bitrate_string = args.bitrate
    ctx.crf_limits = args.crf_limits
    ctx.vmaf_phone_model = args.vmaf_phone_model
    ctx.vmaf_4k_model = args.vmaf_4k_model
    ctx.vmaf_subsample = args.vmaf_subsample
    ctx.vmaf_no_motion = args.vmaf_no_motion
    ctx.auto_accept_autocrop = args.auto_accept_autocrop
    ctx.resolution_preset = args.resolution_preset
    ctx.probe_count = args.probe_count
    ctx.vmaf_reference_display = args.vmaf_reference_display
    ctx.vmaf_probe_speed = args.vmaf_probe_speed
    ctx.crf_map = args.crf_map
    ctx.pin_to_cores = args.dont_pin_to_cores
    ctx.prototype_encoder.niceness = args.niceness
    ctx.vmaf_target_representation = args.vmaf_target_repesentation
    ctx.print_analysis_logs = args.print_analysis_logs
    ctx.poster_url = args.poster_url
    ctx.offload_server = args.offload_server
    ctx.dynamic_vmaf_target = args.dynamic_vmaf_target
    ctx.dynamic_vmaf_target_vbr = args.dynamic_vmaf_target_vbr
    ctx.statically_sized_scenes = args.statically_sized_scenes
    ctx.scene_merge = args.scene_merge
    ctx.args_tune = args.tune
    ctx.denoise_vmaf_ref = args.denoise_vmaf_ref
    ctx.multi_res_pipeline = args.multi_res_pipeline
    ctx.calc_final_vmaf = args.calc_final_vmaf
    ctx.metric_to_target = args.metric_to_target
    ctx.throughput_scaling = args.throughput_scaling

    return ctx, args
